import datetime
import os
import tempfile
from functools import partial
from typing import Any, Literal

import cv2
import matplotlib
import matplotlib as mpl
import matplotlib.colors as mcolors
import matplotlib.patheffects as path_effects
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
from einops import rearrange, repeat
from IPython.display import HTML, Video, display
from jaxtyping import Float, Int
from matplotlib import animation
from matplotlib.colors import Normalize
from matplotlib.patches import Rectangle
from matplotlib.ticker import FixedLocator, FormatStrFormatter
from moviepy import ImageSequenceClip
from scipy.ndimage import center_of_mass
from tqdm.auto import tqdm

from openretina.data_io.hoefling_2024.constants import (
    FRAME_RATE_MODEL,
    MEAN_STD_DICT_74x64,
    pre_normalisation_values_18x16,
)
from openretina.utils.video_analysis import calculate_fft, decompose_kernel, weighted_main_frequency

# Longer animations
matplotlib.rcParams["animation.embed_limit"] = 2**128

# Editable text in PDFs
matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42


def undo_video_normalization(
    video: Float[torch.Tensor, "channels time height width"], values_dict: dict = pre_normalisation_values_18x16
) -> Float[torch.Tensor, "channels time height width"]:
    """
    Undo the normalization of the video.
    """
    video = video.clone()
    video[0] = video[0] * values_dict["channel_0_std"] + values_dict["channel_0_mean"]
    video[1] = video[1] * values_dict["channel_1_std"] + values_dict["channel_1_mean"]

    return video.type(torch.int)


def save_stimulus_to_mp4_video(
    stimulus: Float[torch.Tensor | np.ndarray, "channels time height width"],
    filepath: str,
    fps: int = 5,
    start_at_frame: int = 0,
    apply_undo_video_normalization: bool = False,
) -> None:
    if isinstance(stimulus, torch.Tensor):
        stimulus = stimulus.detach().cpu().numpy()
    assert len(stimulus.shape) == 4
    color_channels = stimulus.shape[0]

    assert filepath.endswith(".mp4")
    # Create a VideoWriter object
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")  # type: ignore
    video = cv2.VideoWriter(filepath, fourcc, fps, (stimulus.shape[3], stimulus.shape[2]))

    # Normalize to uint8
    if apply_undo_video_normalization:
        assert color_channels == 2, "Normalization is only supported for 2 color channels, but {color_channels=}"
        stimulus[0] = stimulus[0] * MEAN_STD_DICT_74x64["channel_0_mean"] + MEAN_STD_DICT_74x64["channel_0_std"]
        stimulus[1] = stimulus[1] * MEAN_STD_DICT_74x64["channel_1_mean"] + MEAN_STD_DICT_74x64["channel_1_std"]
        # Clip to the range of uint8, otherwise there'll be an overflow (-1 will get converted to 255)
        stimulus_uint8 = stimulus.clip(0.0, 255.0).astype(np.uint8)
    else:
        stimulus_norm = stimulus - stimulus.min()
        stimulus_norm = 255 * (stimulus_norm / stimulus_norm.max())
        stimulus_uint8 = stimulus_norm.astype(np.uint8)

    for i in range(start_at_frame, stimulus_uint8.shape[1]):
        # Create an empty 3D array and assign the data from the 4D array
        frame = np.zeros((stimulus_uint8.shape[2], stimulus_uint8.shape[3], 3), dtype=np.uint8)
        if color_channels == 1:
            for c_id in range(3):
                frame[:, :, c_id] = stimulus_uint8[0, i, :, :]
        elif color_channels == 2:
            frame[:, :, 1] = stimulus_uint8[0, i, :, :]  # Green channel
            frame[:, :, 2] = stimulus_uint8[1, i, :, :]  # Red channel
        else:
            frame[:, :, 0] = stimulus_uint8[0, i, :, :]  # Red
            frame[:, :, 1] = stimulus_uint8[1, i, :, :]  # Green
            frame[:, :, 2] = stimulus_uint8[2, i, :, :]  # Blue
        video.write(frame)

    video.release()


def update_video(video, ax, frame):
    """
    Updates the video frame in the given axis.

    To be used in animations.
    """
    ax.clear()

    if video.shape[0] == 1:
        current_frame = video[0, frame].numpy() / video[0, frame].max()

    elif video.shape[0] == 2:
        # Extract the two channels
        green_channel = video[0, frame].numpy()
        purple_channel = video[1, frame].numpy()

        # Create an empty RGB image
        current_frame = np.zeros((*green_channel.shape, 3))

        # Assign green channel to the green color in RGB
        current_frame[:, :, 1] = green_channel

        # Assign purple channel to the blue and red color in RGB (to create purple)
        current_frame[:, :, 0] = purple_channel  # Red
        current_frame[:, :, 2] = purple_channel  # Blue

    else:
        raise NotImplementedError("Only 1 or 2 channels are supported")

    # Display the composite image
    ax.imshow(current_frame, cmap="gray" if video.shape[0] == 1 else None)
    ax.axis("off")


def play_stimulus(video: Float[torch.Tensor, "channels time height width"], normalise: bool = True) -> HTML:
    video = video.detach().cpu()
    if normalise:
        min_val = torch.min(video)
        max_val = torch.max(video)
        video = (video - min_val) / (max_val - min_val)

    fig, ax = plt.subplots()

    update = partial(update_video, video, ax)

    ani = animation.FuncAnimation(fig, update, frames=len(video[0]), interval=50)
    return HTML(ani.to_jshtml())


def play_sample_batch(
    video: Float[torch.Tensor, "channels time height width"],
    responses: Float[torch.Tensor, "time n_neurons"],
    neuron_idx: int | None = 0,
):
    assert video.shape[1] == responses.shape[0], "Movie length and response length must match"

    fig, ax = plt.subplots(1, 2, figsize=(10, 5), gridspec_kw={"width_ratios": [1, 2]})

    def update_trace(frame):
        ax[1].clear()
        ax[1].plot(responses[:frame, neuron_idx])
        ax[1].set_ylim(0, responses[:, neuron_idx].max() * 1.1)
        ax[1].set_xlim(0, len(responses))
        ax[1].set_title(f"Neuron {neuron_idx}")
        ax[1].set_xlabel("Time (frames from start)")
        ax[1].set_ylabel("Response")

    def update_all(frame):
        update_video(video, ax[0], frame)
        update_trace(frame)
        return ax

    ani = animation.FuncAnimation(fig, update_all, frames=len(video[0]), interval=50)
    return HTML(ani.to_jshtml())


def plot_stimulus_composition(
    stimulus: Float[torch.Tensor | np.ndarray, "channels time height width"],
    temporal_trace_ax,
    freq_ax: Any | None,
    spatial_ax,
    lowpass_cutoff: float = 10.0,
    highlight_x_list: list[tuple[int, int]] | None = None,
) -> None:
    if isinstance(stimulus, torch.Tensor):
        stimulus = stimulus.detach().cpu().numpy()
    assert len(stimulus.shape) == 4
    num_color_channels, time_steps, dim_y, dim_x = stimulus.shape

    # guess color channels
    color_array_map = {
        1: ["black"],
        2: ["darkgreen", "darkviolet"],
        3: ["red", "green", "blue"],
    }
    color_channel_names_map = {
        1: ("Grey",),
        2: ("Green", "UV"),
        3: ("Red", "Green", "Blue"),
    }
    color_array = color_array_map[num_color_channels]
    color_channel_names_array = color_channel_names_map[num_color_channels]

    stimulus_time = np.linspace(0, time_steps / FRAME_RATE_MODEL, time_steps)
    weighted_main_freqs = [0.0] * num_color_channels
    temporal_traces_max = 0.0

    temporal_kernels = []
    spatial_kernels = []
    for color_idx in range(num_color_channels):
        temporal, spatial, _ = decompose_kernel(stimulus[color_idx])
        temporal_kernels.append(temporal)
        spatial_kernels.append(spatial)

    # Spatial structure
    spatial_ax.set_title(f"Spatial Component ({'/'.join(color_channel_names_array)})")
    # Create spatial kernel with interleave

    spat = np.concatenate(spatial_kernels, axis=1)
    abs_max = np.max([abs(spat.max()), abs(spat.min())])
    norm = Normalize(vmin=-abs_max, vmax=abs_max)
    spatial_ax.imshow(spat, cmap="RdBu_r", norm=norm)

    # Loop through each kernel and add a black rectangle around it
    x_offset = 0
    for spatial in spatial_kernels:
        kernel_height, kernel_width = spatial.shape
        # kernel_width, kernel_height = spatial.shape

        # Add a black box around the kernel
        rect = Rectangle(
            (x_offset - 0.5, -0.5),
            kernel_width,
            kernel_height,
            linewidth=2,
            edgecolor="black",
            facecolor="none",
            clip_on=False,  # Ensure the rectangle is not clipped by the axes
        )
        spatial_ax.add_patch(rect)

        # Move the x_offset by the width of the current kernel
        x_offset += kernel_width

    # In the low res model the stimulus shape was 18x16 (50 um pixels), for the high-res it is 72x64 (12.5um pixels)
    scale_bar_with = 4 if stimulus.shape[-1] > 20 else 1
    scale_bar = Rectangle(xy=(6, 15), width=scale_bar_with, height=1, color="k", transform=spatial_ax.transData)
    spatial_ax.annotate(
        text="50 µm",
        xy=(6, 14),
    )
    spatial_ax.add_patch(scale_bar)
    spatial_ax.axis("off")

    for color_idx in range(num_color_channels):
        temp = temporal_kernels[color_idx]
        temporal_traces_max = max(temporal_traces_max, np.abs(temp).max())

        # Temporal structure
        temporal_trace_ax.plot(stimulus_time, temp, color=color_array[color_idx])

        if freq_ax is not None:
            fft_freqs, fft_weights = calculate_fft(temp, FRAME_RATE_MODEL, lowpass_cutoff)
            weighted_main_freqs[color_idx] = weighted_main_frequency(fft_freqs, fft_weights)
            freq_ax.plot(fft_freqs, fft_weights, color=color_array[color_idx])

    temporal_trace_ax.set_ylim(-temporal_traces_max, +temporal_traces_max + 1)
    temporal_trace_ax.set_title("Temporal Trace of the Stimulus")
    temporal_trace_ax.set_xlabel("Time [s]")

    if freq_ax is not None:
        freq_ax.set_xlim(0.0, lowpass_cutoff + 1)
        freq_ax.set_xlabel("Frequency [Hz]")
        frequencies_str = "/".join([f"{x:.1f}" for x in weighted_main_freqs])
        freq_ax.set_title(f"Weighted Frequency: {frequencies_str} Hz ({'/'.join(color_channel_names_array)})")

    if highlight_x_list is not None:
        for x_0_idx, x_1_idx in highlight_x_list:
            x_0 = stimulus_time[x_0_idx]
            x_1 = stimulus_time[x_1_idx]
            temporal_trace_ax.fill_betweenx(temporal_trace_ax.get_ylim(), x_0, x_1, color="k", alpha=0.1)


def polar_plot_of_direction_of_motion_responses(
    direction_in_degree: list[int],
    peak_response_per_directions: list[float],
) -> None:
    # Convert directions to radians
    directions_with_peak_response = sorted(
        [(d, v) for d, v in zip(direction_in_degree, peak_response_per_directions, strict=True)]
    )

    # Add the first direction and data point to the end to close the plot
    directions_with_peak_response.append(directions_with_peak_response[0])

    # Create the polar plot
    plt.figure(figsize=(6, 6))
    sorted_directions = [x[0] for x in directions_with_peak_response]
    sorted_data = [x[1] for x in directions_with_peak_response]
    plt.polar(np.deg2rad(sorted_directions), sorted_data, marker="o")

    # Set the direction of the zero point to the top
    plt.gca().set_theta_zero_location("N")  # type: ignore

    # Set the direction of rotation to clockwise
    plt.gca().set_theta_direction(-1)  # type: ignore

    # Set the labels for the directions
    plt.gca().set_xticklabels([f"{x}°" for x in sorted_directions])


class ColorMapper:
    def __init__(self, cmap_name: str, vmin: float = 0, vmax: float = 1):
        self.cmap_name = cmap_name
        self.vmin = vmin
        self.vmax = vmax
        self.cmap = plt.get_cmap(self.cmap_name)
        self.norm = plt.Normalize(vmin=self.vmin, vmax=self.vmax)

    def get_color(self, number):
        color = self.cmap(self.norm(number))
        return color


def plot_vector_field_resp_iso(
    x: np.ndarray,
    y: np.ndarray,
    gradient_dict: np.ndarray,
    resp_dict: np.ndarray,
    normalize_response: bool = False,
    rc_dict: dict[str, Any] = {},
    cmap_lines: str = "hot_r",
    cmap_fill: str = "Greys",
    n_lines=10,
) -> plt.Figure:
    """
    Plots a vector field response with isoresponse lines.

    Parameters:
    x (array-like): The x-coordinates of the grid points.
    y (array-like): The y-coordinates of the grid points.
    gradient_dict (ndarray): A dictionary containing response gradients at grid points.
    resp_dict (ndarray): A dictionary containing responses at grid points.
    normalize_response (bool, optional): If True, normalize the response data. Default is False.
    colour_norm (bool, optional): If True, apply color normalization. Default is False.
    rc_dict (dict, optional): A dictionary of rc settings to use in the plot. Default is an empty dictionary.
    cmap (str, optional): The colormap to use for the plot. Default is "hsv".

    Returns:
    None
    """

    Z = resp_dict.transpose()
    if normalize_response:
        Z = Z / Z.max() * 100
    gradient_grid = gradient_dict[:, 1:-1, 1:-1]
    X, Y = np.meshgrid(x, x)

    # Define levels for isoresponse lines
    levels = np.linspace(Z.min(), Z.max(), n_lines)
    # cm = ColorMapper("cool", vmin=gradient_norm_grid.min(),
    #                  vmax=gradient_norm_grid.max())

    with mpl.rc_context(rc_dict):
        fig = plt.figure()

        # Create a contour plot with isoresponse lines

        plt.contourf(
            X,
            Y,
            Z,
            levels=levels,
            cmap=cmap_fill,
            zorder=200,
        )  # Change cmap to the desired colormap
        cont_lines = plt.contour(X, Y, Z, levels=levels, cmap=cmap_lines, zorder=300)
        plt.gca().clabel(
            cont_lines,
            inline=True,
            fmt="%1.0f",
            levels=list(cont_lines.levels),
            colors="r",
            fontsize=5,
            zorder=400,
        )
        ax = plt.gca()
        ax.set_aspect("equal")

        for i, contrast_green in enumerate(x[1:-1]):
            for j, contrast_uv in enumerate(y[1:-1]):
                unit_vec = gradient_grid[:, i, j] / np.linalg.norm(gradient_grid[:, i, j]) * 0.1
                ax.arrow(
                    contrast_green,
                    contrast_uv,
                    unit_vec[0],
                    unit_vec[1],
                    fill=True,
                    linewidth=0.5,
                    head_width=0.03,
                    color="grey",
                    zorder=300,
                )
        ax.set_xlabel("Green contrast")
        ax.set_ylabel("UV contrast")
        ax.xaxis.set_major_locator(FixedLocator([-1, 0, 1]))
        ax.xaxis.set_major_formatter(FormatStrFormatter("%d"))
        ax.yaxis.set_major_locator(FixedLocator([-1, 0, 1]))
        ax.yaxis.set_major_formatter(FormatStrFormatter("%d"))
        sns.despine()
    return fig


def save_figure(filename: str, folder: str, fig=None) -> None:
    if not os.path.exists(folder):
        os.makedirs(folder, exist_ok=True)
    date = datetime.datetime.now().strftime("%Y-%m-%d")
    filename = f"{date}_{filename}"
    if fig is None:
        fig = plt.gcf()
    fig.savefig(os.path.join(folder, filename))
    plt.close(fig)


def legend_without_duplicate_labels(ax=None) -> None:
    if ax is None:
        ax = plt.gca()
    handles, labels = ax.get_legend_handles_labels()
    unique = [(handle, label) for i, (handle, label) in enumerate(zip(handles, labels)) if label not in labels[:i]]
    ax.legend(*zip(*unique))


def numpy_to_mp4_video(
    video_array: Float[np.ndarray, "time height width 3"],
    save_path: str | os.PathLike | None = None,
    fps: int = 30,
    display_video: bool = True,
    display_width: int = 640,
    display_height: int = 360,
) -> None:
    """
    Convert a NumPy array to mp4 video and optionally display it in a Jupyter Notebook.

    Parameters:
        video_array (np.ndarray): An array of video frames of shape (time, height, width, channels),
                                    where channels must be exactly 3.
        fps (int): Frames per second for the video.
        save_path (str): Path to save the video to. If None, the video is written to a temporary file and
                        just displayed in the notebook.
        display_video (bool): Whether to display the video in the notebook.
        display_width (int): Width of the video iframe in the notebook.
        display_height (int): Height of the video iframe in the notebook.
    """
    assert video_array.ndim == 4, "video_array must have 4 dimensions"
    assert display_video or save_path is not None, "Either display_video or save_path must be provided"

    video_array = prepare_video_for_display(video_array)

    if save_path is not None:
        assert str(save_path).endswith(".mp4"), "save_path must end with '.mp4'"
        file_path = save_path
    else:
        # Create a temporary file
        with tempfile.NamedTemporaryFile(suffix=".mp4", delete=False) as temp_file:
            file_path = temp_file.name

    try:
        # Convert NumPy array to a moviepy ImageSequenceClip
        clip = ImageSequenceClip(list(video_array), fps=fps)

        # Write the video to the file
        clip.write_videofile(file_path, codec="libx264", audio=False, logger=None)

        # Display the video with specified iframe dimensions
        if display_video:
            display(Video(file_path, embed=True, width=display_width, height=display_height))
    finally:
        # Ensure the file is deleted after use if it was temporary
        if save_path is None and os.path.exists(file_path):
            os.remove(file_path)


def stitch_videos(
    video1: Float[np.ndarray, "time height width channels"],
    video2: Float[np.ndarray, "time height width channels"],
    separator_width: int = 50,
    resize_method: Literal["linear", "nearest", "cubic"] = "linear",
) -> np.ndarray:
    """
    Stitch two video arrays side by side with a white band in between.

    Parameters:
        video1 (np.ndarray): Video array of shape (time, height, width, channels).
        video2 (np.ndarray): Video array of shape (time, height, width, channels).
        separator_width (int): Width of the white band separating the videos.
        resize_method (str): Interpolation method for resizing ("linear", "nearest", etc.).

    Returns:
        np.ndarray: Stitched video array of shape (time, max_height, width1 + width2 + band_width, channels).
    """
    assert video1.shape[0] == video2.shape[0], (
        f"Videos must have the same number of frames (time). Got {video1.shape[0]} and {video2.shape[0]}."
    )
    assert video1.shape[3] == video2.shape[3] == 3, (
        f"Videos must have 3 color channels (RGB). Got {video1.shape[3]} and {video2.shape[3]}."
    )
    time, _, _, channels = video1.shape

    max_height = max(video1.shape[1], video2.shape[1])
    bigger_index = 0 if video1.shape == video2.shape else 1 if video1.shape[1] > video2.shape[1] else 2

    interpolation = {
        "nearest": cv2.INTER_NEAREST,
        "linear": cv2.INTER_LINEAR,
        "cubic": cv2.INTER_CUBIC,
    }.get(resize_method, cv2.INTER_LINEAR)

    def resize_video(video, target_height):
        """Resize video frames to the target height while maintaining aspect ratio."""
        scale = target_height / video.shape[1]
        target_width = int(video.shape[2] * scale)
        return np.stack(
            [cv2.resize(frame, (target_width, target_height), interpolation=interpolation) for frame in video], axis=0
        )

    # Resize the videos if necessary
    video1_resized = resize_video(video1, max_height) if bigger_index == 2 else video1
    video2_resized = resize_video(video2, max_height) if bigger_index == 1 else video2

    # Create a white band of the desired width
    band = 255 * np.ones((time, max_height, separator_width, channels), dtype=np.uint8)

    # Concatenate the videos with the band in between
    stitched_video = np.concatenate((video1_resized, band, video2_resized), axis=2)
    return stitched_video


def create_roi_animation(
    roi_mask: Int[np.ndarray, "roi_height roi_width"],
    activity: Float[np.ndarray, "n_neurons time"],
    roi_ids: Int[np.ndarray, " n_neurons"] | None = None,
    cell_types: Int[np.ndarray, " n_neurons"] | None = None,
    min_cell_type: int | None = 1,
    max_cell_type: int | None = 46,
    type_boundaries: list[int] | None = None,
    max_activity: int = 10,
    visualize_ids: bool = False,
    figsize: tuple[int, int] = (8, 8),
) -> Float[np.ndarray, "time fig_height fig_width 3"]:
    """
    Create an animation of neuronal activity and return a numpy tensor of video frames.

    ---
    Parameters:
    roi_mask : np.ndarray
        2D array where background is 1 and neurons are -n (n being neuron index)
    activity : np.ndarray
        n x time array of neuronal activity
    roi_ids : np.ndarray
        1D array indicating the neuron ID for each row in activity (default None)
    cell_types : np.ndarray
        1D array indicating the cell type for each neuron (default None)
    type_boundaries : list[int]
        List of integers indicating boundaries between broad cell types
    max_activity : float
        Maximum activity value for scaling (default 10)
    visualize_ids : bool
        If True, displays neuron IDs on their ROIs
    figsize : tuple
        Figure size in inches (width, height), default (8, 8)

    ---
    Returns:
    np.ndarray
        Numpy array with shape (time, height, width, channels) containing video frames.
    """

    # Get number of neurons and normalize activity
    n_neurons = np.min(roi_mask) * -1
    activity_normalized = np.clip(activity / max_activity, 0, 1)

    # Pre-compute color mapping if cell types are provided
    color_map = {}
    if cell_types is not None:
        max_type = np.max(cell_types) if max_cell_type is None else max_cell_type
        min_type = np.min(cell_types) if min_cell_type is None else min_cell_type

        # Create broad type ranges
        if type_boundaries is None:
            type_boundaries = []
        type_boundaries = sorted([min_type] + type_boundaries + [max_type])

        # For each broad type range, assign a base saturation
        saturations = {}
        for i, (start, end) in enumerate(zip(type_boundaries, type_boundaries[1:])):
            types_in_range = np.where((cell_types >= start) & (cell_types <= end))[0]
            base_saturation = 0.6 + (i * (0.3 / len(type_boundaries)))  # Vary saturation slightly for each broad type
            for type_idx in types_in_range:
                saturations[cell_types[type_idx]] = base_saturation

        # Assign hues based on cell type numbers
        for cell_type in np.unique(cell_types):
            # Map cell type to hue (0-1)
            hue = (cell_type - min_type) / (max_type - min_type)
            saturation = saturations[cell_type]
            color_map[cell_type] = (hue, saturation)

    # Pre-compute centroids for each ROI if we're visualizing IDs
    centroids = {}
    if visualize_ids:
        for i in range(n_neurons):
            if roi_ids is not None and (i + 1) not in roi_ids:
                continue
            neuron_mask = roi_mask == -(i + 1)
            if np.any(neuron_mask):
                centroids[i] = center_of_mass(neuron_mask)

    def render_frame(frame_idx):
        """Render a single frame using matplotlib"""
        # Create figure and axis
        fig, ax = plt.subplots(figsize=figsize)
        plt.close()

        # Create frame image (now in RGB)
        frame_image = np.zeros((*roi_mask.shape, 3))

        # Get the indices to iterate over - either selected ROIs or all neurons
        indices = range(len(roi_ids)) if roi_ids is not None else range(n_neurons)

        # Update each neuron's color based on type and brightness based on activity
        for i in indices:
            # Use roi_ids[i] if provided, otherwise use (i+1) as the ROI ID
            roi_id = roi_ids[i] if roi_ids is not None else (i + 1)
            neuron_mask = roi_mask == -roi_id
            activity_val = activity_normalized[i, frame_idx]

            if cell_types is not None:
                # Get pre-computed hue and saturation for this cell type
                hue, saturation = color_map[cell_types[i]]
                # Convert HSV to RGB
                rgb = mcolors.hsv_to_rgb([hue, saturation, activity_val])
                frame_image[neuron_mask] = rgb
            else:
                # If no cell types provided, use grayscale
                frame_image[neuron_mask] = [activity_val] * 3

        # Display frame
        ax.imshow(frame_image)
        ax.axis("off")

        # Add neuron ID labels if requested
        if visualize_ids:
            for i in [x for x in range(n_neurons) if x in centroids]:
                y, x = centroids[i]
                ax.text(
                    x,
                    y,
                    str(i + 1),
                    color="white",
                    fontsize=8,
                    ha="center",
                    va="center",
                    path_effects=[path_effects.withStroke(linewidth=2, foreground="black")],
                )

        # Redraw canvas to reflect updates
        fig.canvas.draw()

        # Convert matplotlib figure to RGB numpy array
        buffer = fig.canvas.buffer_rgba()
        data = np.asarray(buffer)
        # Convert RGBA to RGB
        data = data[:, :, :3]
        return data

    # Generate all frames
    n_frames = activity.shape[1]
    video_frames = []
    for frame_idx in tqdm(range(n_frames), desc="Generating frames for ROI animation"):
        frame = render_frame(frame_idx)
        video_frames.append(frame)

    # Stack frames into a single numpy array
    video_tensor = np.stack(video_frames, axis=0)
    return video_tensor


def prepare_video_for_display(
    video_array: Float[np.ndarray, "time height width channels"] | Float[np.ndarray, "channels time height width"],
) -> Float[np.ndarray, "time height width 3"]:
    """
    Prepare a video array for display by converting it to uint8 and clipping values to [0, 255].

    Parameters:
    -----------
    video_array : np.ndarray
        Video array of shape (time, height, width, channels) or (channels, time, height, width).

    Returns:
    --------
    np.ndarray
        Video array of shape (time, height, width, channels) with values clipped to [0, 255] and converted to uint8,
        ready for display.
    """
    if np.argmin(video_array.shape) == 0:
        video_array = rearrange(video_array, "channels time height width -> time height width channels")

    if video_array.shape[-1] == 1:
        # Black and white video
        video_array = repeat(video_array, "time height width channel -> time height width (n channel)", n=3)
    if video_array.shape[-1] == 2:
        # Green and UV video
        video_array = np.concatenate((video_array[..., 1][..., None], video_array), axis=-1)

    if video_array.dtype == np.float32 or video_array.dtype == np.float64:
        video_min = video_array.min()
        video_max = video_array.max()
        video_array_normalized = (video_array - video_min) / (video_max - video_min) * 255
        video_array = video_array_normalized.astype(np.uint8)

    # Ensure width and height are even, cannot display in Safari and Firefox otherwise
    height, width = video_array.shape[1:3]
    new_height = height if height % 2 == 0 else height - 1
    new_width = width if width % 2 == 0 else width - 1

    # Crop if necessary
    if (height, width) != (new_height, new_width):
        video_array = video_array[:, :new_height, :new_width, :]

    return video_array.astype(np.uint8)


def display_video(
    video_array: Float[np.ndarray, "time height width channels"] | None,
    video_save_path: str | os.PathLike | None = None,
    fps: int = 30,
    display_width: int = 640,
    display_height: int = 360,
):
    """
    Display a video array or a saved video file.
    This function can either display a video from a numpy array or from a saved video file.
    If a path to an existing video is provided, it will be displayed directly.
    Otherwise, the function will create and display a video from the provided numpy array.

    Args:
        video_array (np.ndarray | None): A 4D numpy array with dimensions (time, height, width, channels)
            representing the video to display. Can be None if video_save_path points to an existing video.
        video_save_path (str | os.PathLike | None): Path to a saved video file. If the file exists,
            it will be displayed directly. If None, video_array must be provided.
        fps (int, optional): Frames per second for video playback. Defaults to 30.
        display_width (int, optional): Width of the displayed video in pixels. Defaults to 640.
        display_height (int, optional): Height of the displayed video in pixels. Defaults to 360.

    Raises:
        AssertionError: If neither a valid video_save_path nor a video_array is provided.

    Examples:
        >>> # Display from numpy array
        >>> video = np.random.rand(100, 480, 640, 3)
        >>> display_video(video, None)
        >>> # Display from saved file
        >>> display_video(None, "path/to/video.mp4")
    """
    if video_save_path is not None and os.path.exists(video_save_path):
        # If save_path already exist, simply display that.
        display(Video(video_save_path, embed=True, width=display_width, height=display_height, mimetype="video/mp4"))
    else:
        assert video_array is not None, (
            "Video array to display must be provided without an already existing saved rendering."
        )
        numpy_to_mp4_video(
            video_array=video_array,
            save_path=video_save_path,
            fps=fps,
            display_video=True,
            display_height=display_height,
            display_width=display_width,
        )



import argparse
import os

import h5py
import numpy as np


def add_parser_arguments(parser: argparse.ArgumentParser):
    parser.description = "Create artificial data to use for testing openretina and as an example to add new datasets"

    parser.add_argument(
        "directory",
        type=str,
        help="Directory to write test data to",
    )

    parser.add_argument(
        "--num-colors",
        type=int,
        default=3,
        help="Color channels in stimulus",
    )

    parser.add_argument(
        "--num-stimuli",
        type=int,
        default=4,
        help="Color channels in stimulus",
    )

    parser.add_argument(
        "--num-sessions",
        type=int,
        default=2,
        help="Number of sessions to generate",
    )

    parser.add_argument(
        "--clip-length",
        type=int,
        default=75,
        help="The number of frames in each clip",
    )


def _generate_response(stimuli: list[np.ndarray], num_neurons: int) -> list[np.ndarray]:
    """Sum over the stimulus in a different area for each neuron and multiply this sum by a scale and add a bias"""
    stimulus_shapes = set((s.shape[0], s.shape[2], s.shape[3]) for s in stimuli)
    assert len(stimulus_shapes) == 1
    stimulus_shape = next(iter(stimulus_shapes))

    num_colors, h, w = stimulus_shape

    # first generate neuron scale, bias and receptive field locations
    scale = np.random.random(num_neurons) + 0.5
    bias = np.random.random(num_neurons)
    color_idc = np.random.randint(0, num_colors, size=num_neurons)
    receptive_field_widths = np.random.randint(2, 6, size=num_neurons)
    h_locations = np.random.randint(0, h - 1, size=num_neurons)
    w_locations = np.random.randint(0, w - 1, size=num_neurons)
    receptive_field_temporal_widths = np.random.randint(3, 7, size=num_neurons)

    responses = []
    for stim in stimuli:
        resp_neurons = []
        for i in range(num_neurons):
            stim_rec_field = stim[
                color_idc[i],
                :,
                h_locations[i] : h_locations[i] + receptive_field_widths[i],
                w_locations[i] : w_locations[i] + receptive_field_widths[i],
            ]
            resp = np.sum(stim_rec_field, axis=-1).sum(axis=-1)
            kernel = np.ones(receptive_field_temporal_widths[i]) / receptive_field_temporal_widths[i]
            resp = np.convolve(resp, kernel, mode="same")
            resp_neurons.append(resp)
        resp_neurons_np = np.stack(resp_neurons)
        resp_neurons_np = resp_neurons_np * scale[:, np.newaxis] + bias[:, np.newaxis]
        responses.append(resp_neurons_np)

    return responses


def write_data_to_directory(directory: str, num_colors: int, num_stimuli: int, num_sessions: int, clip_length: int):
    stimulus_shape_array = [(num_colors, int(t * clip_length), 16, 8) for t in np.arange(70, 70 + num_stimuli)]
    neurons_per_session = np.random.choice(np.arange(20) + 10, size=num_sessions)

    os.makedirs(directory, exist_ok=True)

    stimuli_folder = os.path.join(directory, "stimuli")
    os.makedirs(stimuli_folder, exist_ok=True)

    name_stimulus_list: list[tuple[str, np.ndarray]] = []
    for i, stimulus_shape in enumerate(stimulus_shape_array):
        stim_name = f"random_noise_{i}"
        stim_path = os.path.join(stimuli_folder, stim_name)
        rand_noise = np.random.randn(*stimulus_shape)
        name_stimulus_list.append((stim_name, rand_noise))
        np.save(stim_path, rand_noise, allow_pickle=False)
    print(f"Wrote the following files to the folder {stimuli_folder}: {[x[0] + '.npy' for x in name_stimulus_list]}")

    # generate responses
    for session_id, num_neurons in enumerate(neurons_per_session):
        session_path = os.path.join(directory, f"session_{session_id}.hdf5")
        stimuli = [x[1] for x in name_stimulus_list]
        responses = _generate_response(stimuli, num_neurons)
        with h5py.File(session_path, "w") as f:
            for (stim_name, stim), resp in zip(name_stimulus_list, responses, strict=True):
                resp_name = f"responses_{stim_name}"
                f[resp_name] = resp
            f["session_info/celltypes"] = np.random.choice(40, size=num_neurons)
            print(f"Wrote the following entries to the file {session_path}: {list(f.keys())}")



from openretina.modules.losses.correlation import CorrelationLoss3d
from openretina.modules.losses.mse import MSE3d
from openretina.modules.losses.poisson import PoissonLoss3d

__all__ = ["CorrelationLoss3d", "MSE3d", "PoissonLoss3d"]



from openretina.models import *
model = load_core_readout_from_remote("hoefling_2024_base_low_res", "cpu")
# Visualize a specific channel weight for the first convolutional layer
conv_layer_figure = model.core.plot_weight_visualization(layer=0, in_channel=1, out_channel=0)
# Visualize the readout weights of a neuron for a particular readout session
session_key = model.readout.readout_keys()[0]
readout_figure = model.readout[session_key].plot_weight_for_neuron(5)




#!/usr/bin/env python3

import logging
import os

import hydra
import lightning.pytorch
import torch.utils.data as data
from omegaconf import DictConfig, OmegaConf

from openretina.data_io.base import compute_data_info
from openretina.data_io.cyclers import LongCycler, ShortCycler

log = logging.getLogger(__name__)


@hydra.main(version_base="1.3", config_path="../../configs", config_name="hoefling_2024_core_readout_high_res")
def main(cfg: DictConfig) -> None:
    train_model(cfg)


def train_model(cfg: DictConfig) -> None:
    log.info("Logging full config:")
    log.info(OmegaConf.to_yaml(cfg))

    if cfg.paths.cache_dir is None:
        raise ValueError("Please provide a cache_dir for the data in the config file or as a command line argument.")

    ### Set cache folder
    os.environ["OPENRETINA_CACHE_DIRECTORY"] = cfg.paths.cache_dir

    ### Display log directory for ease of access
    log.info(f"Saving run logs at: {cfg.paths.output_dir}")

    ### Import data
    movies_dict = hydra.utils.call(cfg.data_io.stimuli)
    neuron_data_dict = hydra.utils.call(cfg.data_io.responses)

    if cfg.check_stimuli_responses_match:
        for session, neuron_data in neuron_data_dict.items():
            neuron_data.check_matching_stimulus(movies_dict[session])

    dataloaders = hydra.utils.instantiate(
        cfg.dataloader,
        neuron_data_dictionary=neuron_data_dict,
        movies_dictionary=movies_dict,
    )

    data_info = compute_data_info(neuron_data_dict, movies_dict)

    train_loader = data.DataLoader(
        LongCycler(dataloaders["train"], shuffle=True), batch_size=None, num_workers=0, pin_memory=True
    )
    valid_loader = ShortCycler(dataloaders["validation"])

    if cfg.seed is not None:
        lightning.pytorch.seed_everything(cfg.seed)

    ### Model init
    # Assign missing n_neurons_dict to the model
    cfg.model.n_neurons_dict = data_info["n_neurons_dict"]

    log.info(f"Instantiating model <{cfg.model._target_}>")
    model = hydra.utils.instantiate(cfg.model, data_info=data_info)

    ### Logging
    log.info("Instantiating loggers...")
    logger_array = []
    for _, logger_params in cfg.logger.items():
        logger = hydra.utils.instantiate(logger_params)
        logger_array.append(logger)

    ### Callbacks
    log.info("Instantiating callbacks...")
    callbacks = [
        hydra.utils.instantiate(callback_params) for callback_params in cfg.get("training_callbacks", {}).values()
    ]

    ### Trainer init
    log.info(f"Instantiating trainer <{cfg.trainer._target_}>")
    trainer: lightning.Trainer = hydra.utils.instantiate(cfg.trainer, logger=logger_array, callbacks=callbacks)
    trainer.fit(model=model, train_dataloaders=train_loader, val_dataloaders=valid_loader)

    ### Testing
    log.info("Starting testing!")
    short_cyclers = [(n, ShortCycler(dl)) for n, dl in dataloaders.items()]
    dataloader_mapping = {f"DataLoader {i}": x[0] for i, x in enumerate(short_cyclers)}
    log.info(f"Dataloader mapping: {dataloader_mapping}")
    trainer.test(model, dataloaders=[c for _, c in short_cyclers], ckpt_path="best")


if __name__ == "__main__":
    main()



import numpy as np
from einops import rearrange, repeat
from jaxtyping import Float

from openretina.eval.metrics import correlation_numpy


def oracle_corr_jackknife(
    repeated_responses: Float[np.ndarray, "frames repeats neurons"],
) -> tuple[Float[np.ndarray, " neurons"], Float[np.ndarray, " frames repeats neurons"]]:
    """
    Adapted from neuralpredictors.
    Compute the oracle correlations per neuron by averaging over repeated responses in a leave one out fashion.
    Note that oracle_corr_jackknife underestimates the true oracle correlation.

    Args:
        repeated_responses (array-like): numpy array with shape (images/time, repeats, neuron responses).

    Returns:
        tuple: A tuple containing:
            - oracle_score (array): Oracle correlation for each neuron
            - oracle (array): Oracle responses for each neuron
    """

    loo_oracles = []
    for response_t in repeated_responses:
        num_repeats = response_t.shape[0]
        # Compute the oracle by averaging over all repeats except the current one
        # (add all, subtract current, divide by num_repeats - 1)
        oracle_t = (response_t.sum(axis=0, keepdims=True) - response_t) / (num_repeats - 1)
        oracle_t = np.nan_to_num(oracle_t)
        loo_oracles.append(oracle_t)
    oracle = np.stack(loo_oracles)

    oracle_score = correlation_numpy(
        rearrange(repeated_responses, "t r n -> (t r) n"),
        rearrange(oracle, "t r n -> (t r) n", t=repeated_responses.shape[0]),
        axis=0,
    )

    return oracle_score, oracle


def global_mean_oracle(
    responses: Float[np.ndarray, "frames repeats neurons"] | Float[np.ndarray, "frames neurons"],
) -> Float[np.ndarray, " neurons"]:
    """
    Compute the oracle correlation between each neuron's response and the global mean response.

    The global mean oracle correlation represents how well each neuron's activity can be predicted
    by the average response across all neurons at each time point.

    Args:
        responses (np.ndarray): Neural responses array. Can be either:
            - 3D array of shape (frames, repeats, neurons)
            - 2D array of shape (frames, neurons) which will be treated as single repeat
        return_oracle (bool, optional): If True, returns both correlation values and oracle responses.
            Defaults to False.

    Returns:
        - 1D array of shape (neurons,) containing correlation values for each neuron

    Note:
        The function automatically handles single-repeat data by adding a singleton dimension.
    """
    if responses.ndim == 2:
        responses = responses[:, None, :]

    global_mean_response = repeat(responses.mean(axis=2, keepdims=True), "t r _ -> t r n", n=responses.shape[2])

    oracle_mean_corr = correlation_numpy(
        rearrange(responses, "t r n -> (t r) n"),
        rearrange(global_mean_response, "t r n -> (t r) n", t=responses.shape[0]),
        axis=0,
    )

    return oracle_mean_corr



from contextlib import contextmanager

import torch
from lightning.pytorch.callbacks import Callback


@contextmanager
def eval_state(model: torch.nn.Module):
    training_status = model.training

    try:
        model.eval()
        yield model
    finally:
        model.train(training_status)


class OptimizerResetCallback(Callback):
    def __init__(self):
        super().__init__()
        self.prev_lr = None  # This will store the previous learning rate

    def on_validation_end(self, trainer, pl_module):
        # Get the current learning rate from the optimizer
        optims = pl_module.optimizers()
        try:
            optim = optims[0]
        except:  # noqa
            optim = optims
        current_lr = optim.param_groups[0]["lr"]

        # Compare with the previous learning rate
        if self.prev_lr is not None and current_lr < self.prev_lr:
            print(f"Learning rate decreased from {self.prev_lr} to {current_lr}. Resetting optimizer.")
            # Reset the optimizer if the learning rate has decreased
            params_dict = optim.param_groups[0]
            # below could be written shorter
            new_optimizer = torch.optim.AdamW(
                pl_module.parameters(),
                lr=current_lr,
                betas=params_dict["betas"],
                eps=params_dict["eps"],
                weight_decay=params_dict["weight_decay"],
                amsgrad=params_dict["amsgrad"],
                maximize=params_dict["maximize"],
                foreach=params_dict["foreach"],
                capturable=params_dict["capturable"],
                differentiable=params_dict["differentiable"],
                fused=params_dict["fused"],
            )
            trainer.optimizers = [new_optimizer]  # Replace the optimizer in the trainer

        self.prev_lr = current_lr



import functools
from abc import ABC, abstractmethod
from collections import OrderedDict

import torch


class ResponseReducer(ABC):
    def __init__(self, axis: int = 0):
        self._axis = axis

    @abstractmethod
    def forward(self, responses: torch.Tensor) -> torch.Tensor:
        pass


class MeanReducer(ResponseReducer):
    def __init__(self, axis: int = 0):
        super().__init__(axis)

    def forward(self, responses: torch.Tensor) -> torch.Tensor:
        return torch.mean(responses, dim=self._axis)


class SliceMeanReducer(ResponseReducer):
    def __init__(self, axis: int, start: int, length: int):
        super().__init__(axis)
        self.start = start
        self.length = length

    def forward(self, responses: torch.Tensor) -> torch.Tensor:
        narrowed_responses = torch.narrow(responses, self._axis, self.start, self.length)
        return torch.mean(narrowed_responses, dim=self._axis)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._axis=} {self.start=}, {self.length=})"


class AbstractObjective(ABC):
    def __init__(self, model, data_key: str | None):
        self._model = model
        self._data_key = data_key

    def model_forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        if self._data_key is not None:
            responses = self._model(stimulus, data_key=self._data_key)
        else:
            responses = self._model(stimulus)
        # squeeze batch dimension
        return responses.squeeze(0)

    @abstractmethod
    def forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        pass


class IncreaseObjective(AbstractObjective):
    def __init__(self, model, neuron_indices: list[int] | int, data_key: str | None, response_reducer: ResponseReducer):
        super().__init__(model, data_key)
        self._neuron_indices = [neuron_indices] if isinstance(neuron_indices, int) else neuron_indices
        self._response_reducer = response_reducer

    def forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        responses = self.model_forward(stimulus)
        # responses.shape = (time, neuron)
        selected_responses = responses[:, self._neuron_indices]
        mean_response = selected_responses.mean(dim=-1)
        # average over time dimension
        single_score = self._response_reducer.forward(mean_response)
        return single_score


class _ModuleHook:
    """This structure is used to store the output of this module.
    This is achieved by the hook_fn as a forward hook which is called on every forward
    path of this module.
    We then store the output tensor which can then be used downstream, e.g. to optimize an MEI.
    Pass name to the initializer to get verbose output.
    """

    def __init__(self, module: torch.nn.Module, name: str | None = None):
        self.hook = module.register_forward_hook(self.hook_fn)
        self.module = module
        self.module_output_tensor: torch.Tensor | None = None
        self.name = name

    def hook_fn(self, module: torch.nn.Module, input_: tuple, output: torch.Tensor):
        assert self.module is module
        self.module_output_tensor = output
        if self.name is not None:
            print(f"Module hook {self.name} hooked up {type(self.module)=} {self.module_output_tensor.shape=}")

    def close(self) -> None:
        self.hook.remove()


class InnerNeuronVisualizationObjective(AbstractObjective):
    def __init__(self, model, data_key: str | None, response_reducer: ResponseReducer):
        super().__init__(model, data_key)
        self.features_dict = self.hook_model(model)
        self._response_reducer = response_reducer
        self.layer_name = ""
        self.channel_id = -1

    def set_layer_channel(self, layer: str, channel: int) -> None:
        if layer not in self.features_dict:
            raise ValueError(f"{layer=} not in features {self.features_dict.keys=}")
        self.layer_name = layer
        self.channel_id = channel

    @functools.lru_cache
    def get_output_shape_for_layer(self, layer_name: str, stimulus_shape: tuple[int, ...]) -> tuple[int, ...] | None:
        if layer_name not in self.features_dict:
            return None

        with torch.no_grad():
            model_device = next(self._model.parameters()).device
            stimulus = torch.rand(stimulus_shape, device=model_device, requires_grad=False)
            self.model_forward(stimulus)
            output_tensor = self.features_dict[layer_name].module_output_tensor
            if output_tensor is None:
                output_shape = None
            else:
                output_shape = output_tensor.shape
        return output_shape

    def forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        if len(self.layer_name) == 0 or self.channel_id < 0:
            raise ValueError("layer_name and channel_id not set. Please call `set_layer_channel`.")
        # The forward call is just used to update the model hooks
        self.model_forward(stimulus)
        module_hook = self.features_dict[self.layer_name]
        if module_hook.module_output_tensor is None:
            raise ValueError(f"Module hook output tensor is None for layer {self.layer_name}: {module_hook=}")
        output = module_hook.module_output_tensor[:, self.channel_id]
        y, x = output.shape[-2:]
        middle_neuron_trace_batch = output[..., y // 2, x // 2]
        middle_neuron_trace = middle_neuron_trace_batch.mean(axis=0)
        loss = self._response_reducer.forward(middle_neuron_trace)
        return loss

    @staticmethod
    def hook_model(model) -> OrderedDict:
        features = OrderedDict()

        # recursive hooking function
        def hook_layers(net, prefix: tuple[str, ...] = ()) -> None:
            if hasattr(net, "_modules"):
                for name, layer in net._modules.items():
                    if layer is not None:
                        new_prefix_tuple = prefix + (name,)
                        features_key = "_".join(new_prefix_tuple)
                        features[features_key] = _ModuleHook(layer)
                        hook_layers(layer, prefix=new_prefix_tuple)

        hook_layers(model)

        return features


class ContrastiveNeuronObjective(AbstractObjective):
    """Objective described in [Most discriminative stimuli for functional cell type clustering]
    (https://openreview.net/forum?id=9W6KaAcYlr)"""

    def __init__(
        self,
        model,
        on_cluster_idc: list[int],
        off_cluster_idc_list: list[list[int]],
        data_key: str | None,
        response_reducer: ResponseReducer,
        temperature: float = 1.6,
    ):
        super().__init__(model, data_key)
        self._on_cluster_idc = on_cluster_idc
        self._off_cluster_idc_list = off_cluster_idc_list
        self._response_reducer = response_reducer
        self._temperature = temperature

    @staticmethod
    def contrastive_objective(on_score: torch.Tensor, all_scores: torch.Tensor, temperature: float) -> torch.Tensor:
        t = temperature
        obj = (
            (1 / t) * on_score
            - torch.logsumexp((1 / t) * all_scores, dim=0)
            + torch.log(torch.tensor(all_scores.size(0)))
        )
        return obj

    def forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        responses = self.model_forward(stimulus)
        score_per_neuron = self._response_reducer.forward(responses)

        on_score = score_per_neuron[self._on_cluster_idc].mean()
        off_scores = [score_per_neuron[idc].mean() for idc in self._off_cluster_idc_list]
        obj = self.contrastive_objective(
            on_score,
            torch.stack([on_score] + off_scores),
            self._temperature,
        )
        return obj



import sys
from typing import Optional


class OptimizationStopper:
    def __init__(self, max_iterations: Optional[int]):
        if max_iterations is None:
            self.max_iterations = sys.maxsize
        else:
            self.max_iterations = max_iterations

    def early_stop(self, loss: float) -> bool:
        return False


class EarlyStopper(OptimizationStopper):
    def __init__(self, max_iterations: Optional[int] = None, patience: int = 1, min_delta: float = 0.0):
        super().__init__(max_iterations)
        self._patience = patience
        self._min_delta = min_delta
        self._counter = 0
        self._min_loss = float("inf")

    def early_stop(self, loss: float) -> bool:
        if loss < self._min_loss:
            self._min_loss = loss
            self._counter = 0
        elif loss > (self._min_loss + self._min_delta):
            self._counter += 1
            if self._counter >= self._patience:
                return True
        return False



import torch.nn


class MSE3d(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, output, target):
        lag = target.size(1) - output.size(1)
        loss = output - target[:, lag:, :]
        return loss.pow(2).sum(dim=-1).sum()



import bisect
import collections
from collections import namedtuple
from typing import Any, List, Literal, Optional

import numpy as np
import torch
from jaxtyping import Float
from torch.utils.data import DataLoader, Dataset, Sampler
from tqdm.auto import tqdm

from openretina.data_io.base import MoviesTrainTestSplit, ResponsesTrainTestSplit

DataPoint = namedtuple("DataPoint", ["inputs", "targets"])


class MovieDataSet(Dataset):
    """
    A dataset class for handling movie data and corresponding neural responses.

    Args:
        movies (Float[np.ndarray | torch.Tensor, "n_channels n_frames h w"]): The movie data.
        responses (Float[np.ndarray, "n_frames n_neurons"]): The neural responses.
        roi_ids (Optional[Float[np.ndarray, " n_neurons"]]): A list of ROI IDs.
        roi_coords (Optional[Float[np.ndarray, "n_neurons 2"]]): A list of ROI coordinates.
        group_assignment (Optional[Float[np.ndarray, " n_neurons"]]): A list of group assignments (cell types).
        split (Literal["train", "validation", "val", "test"]):
                                                    The data split, either "train", "validation", "val", or "test".
        chunk_size (int): The size of the chunks to split the data into.

    Attributes:
        samples (tuple): A tuple containing movie data and neural responses.
        test_responses_by_trial (Optional[Dict[str, Any]]):
                                                A dictionary containing test responses by trial (only for test split).
        roi_ids (Optional[Float[np.ndarray, " n_neurons"]]): A list of region of interest (ROI) IDs.
        chunk_size (int): The size of the chunks to split the data into.
        mean_response (torch.Tensor): The mean response per neuron.
        group_assignment (Optional[Float[np.ndarray, " n_neurons"]]): A list of group assignments.
        roi_coords (Optional[Float[np.ndarray, "n_neurons 2"]]): A list of ROI coordinates.

    Methods:
        __getitem__(idx): Returns a DataPoint object for the given index or slice.
        movies: Returns the movie data.
        responses: Returns the neural responses.
        __len__(): Returns the number of chunks of clips and responses used for training.
        __str__(): Returns a string representation of the dataset.
        __repr__(): Returns a string representation of the dataset.
    """

    def __init__(
        self,
        movies: Float[np.ndarray | torch.Tensor, "n_channels n_frames h w"],
        responses: Float[np.ndarray | torch.Tensor, "n_frames n_neurons"],
        roi_ids: Float[np.ndarray, " n_neurons"] | None,
        roi_coords: Float[np.ndarray, "n_neurons 2"] | None,
        group_assignment: Float[np.ndarray, " n_neurons"] | None,
        split: str | Literal["train", "validation", "val", "test"],
        chunk_size: int,
    ):
        # Will only be a dictionary for certain types of datasets, i.e. Hoefling 2022
        if split == "test" and isinstance(responses, dict):
            self.samples: tuple = movies, responses["avg"]
            self.test_responses_by_trial = responses["by_trial"]
            self.roi_ids = roi_ids
        else:
            self.samples = movies, responses

        self.chunk_size = chunk_size
        # Calculate the mean response per neuron (used for bias init in the model)
        self.mean_response = torch.mean(torch.Tensor(self.samples[1]), dim=0)
        self.group_assignment = group_assignment
        self.roi_coords = roi_coords

    def __getitem__(self, idx: int | slice) -> DataPoint:
        if isinstance(idx, slice):
            return DataPoint(*[self.samples[0][:, idx, ...], self.samples[1][idx, ...]])
        else:
            return DataPoint(
                *[
                    self.samples[0][:, idx : idx + self.chunk_size, ...],
                    self.samples[1][idx : idx + self.chunk_size, ...],
                ]
            )

    @property
    def movies(self):
        return self.samples[0]

    @property
    def responses(self):
        return self.samples[1]

    def __len__(self) -> int:
        # Returns the number of chunks of clips and responses used for training
        return self.samples[1].shape[0] // self.chunk_size

    def __str__(self) -> str:
        return (
            f"MovieDataSet with {self.samples[1].shape[1]} neuron responses "
            f"to a movie of shape {list(self.samples[0].shape)}."
        )

    def __repr__(self) -> str:
        return str(self)


def generate_movie_splits(
    movie_train,
    movie_test: dict[str, np.ndarray],
    val_clip_idc: list[int],
    num_clips: int,
    clip_length: int,
) -> tuple[torch.Tensor, torch.Tensor, dict[str, torch.Tensor]]:
    movie_train = torch.tensor(movie_train, dtype=torch.float)
    movie_test_dict = {n: torch.tensor(movie, dtype=torch.float) for n, movie in movie_test.items()}

    channels, _, px_y, px_x = movie_train.shape

    # Prepare validation movie data
    movie_val = torch.zeros((channels, len(val_clip_idc) * clip_length, px_y, px_x), dtype=torch.float)
    for i, idx in enumerate(val_clip_idc):
        movie_val[:, i * clip_length : (i + 1) * clip_length, ...] = movie_train[
            :, idx * clip_length : (idx + 1) * clip_length, ...
        ]

    # Create a boolean mask to indicate which clips are not part of the validation set
    mask = np.ones(num_clips, dtype=bool)
    mask[val_clip_idc] = False
    train_clip_idx = np.arange(num_clips)[mask]

    movie_train_subset = torch.cat(
        [movie_train[:, i * clip_length : (i + 1) * clip_length] for i in train_clip_idx],
        dim=1,
    )

    return movie_train_subset, movie_val, movie_test_dict


class MovieSampler(Sampler):
    """
    A custom sampler for selecting movie frames for training, validation, or testing.

    Args:
        start_indices (list[int]): List of starting indices for the movie sections to select.
        split (Literal["train", "validation", "val", "test"]): The type of data split.
        chunk_size (int): The size of each contiguous chunk of frames to select.
        movie_length (int): The total length of the movie.
        scene_length (Optional[int], optional): The length of each scene, if the movie is divided in any scenes.
                                                Defaults to None.
        allow_over_boundaries (bool, optional): Whether to allow selected chunks to go over scene boundaries.
                                                Defaults to False.

    Attributes:
        indices (list[int]): The starting indices for the movie sections to sample.
        split (str): The type of data split.
        chunk_size (int): The size of each chunk of frames.
        movie_length (int): The total length of the movie.
        scene_length (int): The length of each scene, if the movie is made up of scenes.
        allow_over_boundaries (bool): Whether to allow chunks to go over scene boundaries.

    Methods:
        __iter__(): Returns an iterator over the sampled indices.
        __len__(): Returns the number of starting indices (which will corresponds to the number of sampled clips).
    """

    def __init__(
        self,
        start_indices: list[int],
        split: str | Literal["train", "validation", "val", "test"],
        chunk_size: int,
        movie_length: int,
        scene_length: int,
        allow_over_boundaries: bool = False,
    ):
        super().__init__()
        self.indices = start_indices
        self.split = split
        self.chunk_size = chunk_size
        self.movie_length = movie_length
        self.scene_length = scene_length
        self.allow_over_boundaries = allow_over_boundaries

    def __iter__(self):
        if self.split == "train" and (self.scene_length != self.chunk_size):
            if self.allow_over_boundaries:
                shifts = np.random.randint(0, self.chunk_size, len(self.indices))
                # apply shifts while making sure we do not exceed the movie length
                shifted_indices = np.minimum(self.indices + shifts, self.movie_length - self.chunk_size)
            else:
                shifted_indices = gen_shifts_with_boundaries(
                    np.arange(0, self.movie_length + 1, self.scene_length),
                    self.indices,
                    self.chunk_size,
                )
            # Shuffle the indices
            indices_shuffling = np.random.permutation(len(self.indices))
        else:
            shifted_indices = self.indices
            indices_shuffling = np.arange(len(self.indices))

        return iter(np.array(shifted_indices)[indices_shuffling])

    def __len__(self) -> int:
        return len(self.indices)


def gen_shifts_with_boundaries(
    clip_bounds: list[int] | np.ndarray, start_indices: list[int] | np.ndarray, clip_chunk_size: int = 50
) -> list[int]:
    """
    Generate shifted indices based on clip bounds and start indices.
    Assumes that the original start indices are already within the clip bounds.
    If they are not, it changes the overflowing indexes to respect the closest bound.

    Args:
        clip_bounds (list): A list of clip bounds.
        start_indices (list): A list of start indices.
        clip_chunk_size (int, optional): The size of each clip chunk. Defaults to 50.

    Returns:
        list: A list of shifted indices.

    """

    def get_next_bound(value, bounds):
        insertion_index = bisect.bisect_right(bounds, value)
        return bounds[min(insertion_index, len(bounds) - 1)]

    shifted_indices = []
    shifts = np.random.randint(0, clip_chunk_size // 2, len(start_indices))

    for i, start_idx in enumerate(start_indices):
        next_bound = get_next_bound(start_idx, clip_bounds)
        if start_idx + shifts[i] + clip_chunk_size < next_bound:
            shifted_indices.append(start_idx + shifts[i])
        elif start_idx + clip_chunk_size > next_bound:
            shifted_indices.append(next_bound - clip_chunk_size)
        else:
            shifted_indices.append(start_idx)

    # Ensure we do not exceed the movie length when allowing over boundaries
    if shifted_indices[-1] + clip_chunk_size > clip_bounds[-1]:
        shifted_indices[-1] = clip_bounds[-1] - clip_chunk_size
    return shifted_indices


def handle_missing_start_indices(
    movie_length: int, chunk_size: int | None, scene_length: int | None, split: str
) -> list[int]:
    """
    Handle missing start indices for different splits of the dataset.

    Parameters:
    movies (np.ndarray or torch.Tensor): The movies data, as an array.
    chunk_size (int or None): The size of each chunk for training split. Required if split is "train".
    scene_length (int or None): The length of each scene. Required if split is "validation" or "val".
    split (str): The type of split, one of "train", "validation", "val", or "test".

    Returns:
    dict or list: The generated or provided start indices for each movie.

    Raises:
    AssertionError: If chunk_size is not provided for training split when start_indices is None.
    AssertionError: If scene_length is not provided for validation split when start_indices is None.
    NotImplementedError: If start_indices is None and split is not one of "train", "validation", "val", or "test".
    """

    if split == "train":
        assert chunk_size is not None, "Chunk size or start indices must be provided for training."
        interval = chunk_size
    elif split in {"validation", "val"}:
        assert scene_length is not None, "Scene length or start indices must be provided for validation."
        interval = scene_length
    elif split == "test":
        interval = movie_length
    else:
        raise NotImplementedError("Start indices could not be recovered.")

    return np.arange(0, movie_length, interval).tolist()  # type: ignore


def get_movie_dataloader(
    movie: Float[np.ndarray | torch.Tensor, "n_channels n_frames h w"],
    responses: Float[np.ndarray | torch.Tensor, "n_frames n_neurons"],
    *,
    split: str | Literal["train", "validation", "val", "test"],
    scene_length: int,
    chunk_size: int,
    batch_size: int,
    start_indices: list[int] | None = None,
    roi_ids: Float[np.ndarray, " n_neurons"] | None = None,
    roi_coords: Float[np.ndarray, "n_neurons 2"] | None = None,
    group_assignment: Float[np.ndarray, " n_neurons"] | None = None,
    drop_last: bool = True,
    allow_over_boundaries: bool = True,
    **kwargs,
) -> DataLoader:
    """
    Create a DataLoader for processing movie data and associated responses.
    This function prepares the dataset and sampler for training or evaluation based on the specified parameters.

    Args:
        movie (Float[np.ndarray | torch.Tensor, "n_channels n_frames h w"]):
            The movie data represented as a multi-dimensional array or tensor.
        responses (Float[np.ndarray, "n_frames n_neurons"]):
            The responses corresponding to the frames of the movie.
        split (str | Literal["train", "validation", "val", "test"]):
            The dataset split to use (train, validation, or test).
        scene_length (int):
            The length of the scene to be processed.
        chunk_size (int):
            The size of each chunk to be extracted from the movie.
        batch_size (int):
            The number of samples per batch.
        start_indices (list[int] | None, optional):
            The starting indices for each chunk. If None, will be computed.
        roi_ids (Float[np.ndarray, " n_neurons"] | None, optional):
            The region of interest IDs. If None, will not be used.
        roi_coords (Float[np.ndarray, "n_neurons 2"] | None, optional):
            The coordinates of the regions of interest. If None, will not be used.
        group_assignment (Float[np.ndarray, " n_neurons"] | None, optional):
            The group assignments (cell types) for the neurons. If None, will not be used.
        drop_last (bool, optional):
            Whether to drop the last incomplete batch. Defaults to True.
        allow_over_boundaries (bool, optional):
            Whether to allow chunks that exceed the scene boundaries. Defaults to True.
        **kwargs:
            Additional keyword arguments for the DataLoader.

    Returns:
        DataLoader:
            A DataLoader instance configured with the specified dataset and sampler.

    Raises:
        ValueError:
            If `allow_over_boundaries` is False and `chunk_size` exceeds `scene_length` during training.
    """
    if isinstance(responses, torch.Tensor) and bool(torch.isnan(responses).any()):
        print("Nans in responses, skipping this dataloader")
        return  # type: ignore

    if not allow_over_boundaries and split == "train" and chunk_size > scene_length:
        raise ValueError("Clip chunk size must be smaller than scene length to not exceed clip bounds during training.")

    if start_indices is None:
        start_indices = handle_missing_start_indices(movie.shape[1], chunk_size, scene_length, split)
    dataset = MovieDataSet(movie, responses, roi_ids, roi_coords, group_assignment, split, chunk_size)
    sampler = MovieSampler(
        start_indices,
        split,
        chunk_size,
        movie_length=movie.shape[1],
        scene_length=scene_length,
        allow_over_boundaries=allow_over_boundaries,
    )

    return DataLoader(
        dataset, sampler=sampler, batch_size=batch_size, drop_last=split == "train" and drop_last, **kwargs
    )


class NeuronDataSplit:
    def __init__(
        self,
        responses: ResponsesTrainTestSplit,
        val_clip_idx: List[int],
        num_clips: int,
        clip_length: int,
        key: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the NeuronData object.
        Boilerplate class to compute and store neuron data train/test/validation splits before feeding into a dataloader

        Args:
            key (dict): The key information for the neuron data,
                        includes date, exp_num, experimenter, field_id, stim_id.
            responses (ResponsesTrainTestSplit): The train and test responses of neurons.
            val_clip_idx (List[int]): The indices of validation clips.
            num_clips (int): The number of clips.
            clip_length (int): The length of each clip.
            key (dict, optional): Additional key information.
        """
        self.neural_responses = responses
        self.num_neurons = self.neural_responses.n_neurons
        self.key = key
        self.roi_coords = ()
        self.clip_length = clip_length
        self.num_clips = num_clips
        self.val_clip_idx = val_clip_idx

        # Transpose the responses to have the shape (n_timepoints, n_neurons)
        self.responses_train_and_val = self.neural_responses.train.T

        self.responses_train, self.responses_val = self.split_data_train_val()
        self.test_responses_by_trial = np.array([])  # Added for compatibility with Hoefling et al., 2024

    def split_data_train_val(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Compute validation responses and updated train responses stripped from validation clips.
        Can deal with unsorted validation clip indices, and parallels the way movie validation clips are handled.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The updated train and validation responses.
        """
        # Initialise validation responses
        base_movie_sorting = np.arange(self.num_clips)

        validation_mask = np.ones_like(self.responses_train_and_val, dtype=bool)
        responses_val = np.zeros([len(self.val_clip_idx) * self.clip_length, self.num_neurons])

        # Compute validation responses and remove sections from training responses
        for i, ind1 in enumerate(self.val_clip_idx):
            grab_index = base_movie_sorting[ind1]
            responses_val[i * self.clip_length : (i + 1) * self.clip_length, :] = self.responses_train_and_val[
                grab_index * self.clip_length : (grab_index + 1) * self.clip_length,
                :,
            ]
            validation_mask[
                (grab_index * self.clip_length) : (grab_index + 1) * self.clip_length,
                :,
            ] = False

        responses_train = self.responses_train_and_val[validation_mask].reshape(-1, self.num_neurons)

        return responses_train, responses_val

    @property
    def response_dict(self) -> dict:
        """
        Create and return a dictionary of neural responses for train, validation, and test datasets.
        """
        return {
            "train": torch.tensor(self.responses_train, dtype=torch.float),
            "validation": torch.tensor(self.responses_val, dtype=torch.float),
            "test": {
                "avg": self.response_dict_test,
                "by_trial": torch.tensor(self.test_responses_by_trial, dtype=torch.float),
            },
        }

    @property
    def response_dict_test(self) -> dict[str, torch.Tensor]:
        return {name: torch.tensor(responses.T) for name, responses in self.neural_responses.test_dict.items()}


def get_movie_splits(
    movie_train,
    movie_test: dict[str, np.ndarray],
    val_clip_idx: list[int],
    num_clips: int,
    clip_length: int,
) -> dict[str, Any]:
    movie_train_subset, movie_val, movie_test_tensor = generate_movie_splits(
        movie_train, movie_test, val_clip_idx, num_clips, clip_length
    )

    movies = {
        "train": movie_train_subset,
        "validation": movie_val,
        "test_dict": movie_test_tensor,
        "val_clip_idx": val_clip_idx,
    }

    return movies


def multiple_movies_dataloaders(
    neuron_data_dictionary: dict[str, ResponsesTrainTestSplit],
    movies_dictionary: dict[str, MoviesTrainTestSplit],
    train_chunk_size: int = 50,
    batch_size: int = 32,
    seed: int = 42,
    clip_length: int = 100,
    num_val_clips: int = 10,
    val_clip_indices: list[int] | None = None,
    allow_over_boundaries: bool = True,
) -> dict[str, dict[str, DataLoader]]:
    """
    Create multiple dataloaders for training, validation, and testing from given neuron and movie data.
    This function ensures that the neuron data and movie data are aligned and generates dataloaders for each session.
    It does not make assumptions about the movies in different sessions to be the same, the same length, composed
    of the same clips or in the same order.

    Args:
        neuron_data_dictionary (dict[str, ResponsesTrainTestSplit]):
            A dictionary containing neuron response data split for training and testing.
        movies_dictionary (dict[str, MoviesTrainTestSplit]):
            A dictionary containing movie data split for training and testing.
        train_chunk_size (int, optional):
            The size of the chunks for training data. Defaults to 50.
        batch_size (int, optional):
            The number of samples per batch. Defaults to 32.
        seed (int, optional):
            The random seed for reproducibility. Defaults to 42.
        clip_length (int, optional):
            The length of each clip. Defaults to 100.
        num_val_clips (int, optional):
            The number of validation clips to draw. Defaults to 10.
        val_clip_indices (list[int], optional): The indices of validation clips to use. If provided, num_val_clips is
                                                ignored. Defaults to None.
        allow_over_boundaries (bool, optional):  Whether to allow selected chunks to go over scene boundaries.

    Returns:
        dict:
            A dictionary containing dataloaders for training, validation, and testing for each session.

    Raises:
        AssertionError:
            If the keys of neuron_data_dictionary and movies_dictionary do not match exactly.
    """
    assert set(neuron_data_dictionary.keys()) == set(movies_dictionary.keys()), (
        "The keys of neuron_data_dictionary and movies_dictionary should match exactly."
    )

    # Initialise dataloaders
    dataloaders: dict[str, Any] = collections.defaultdict(dict)

    for session_key, session_data in tqdm(neuron_data_dictionary.items(), desc="Creating movie dataloaders"):
        # Extract all data related to the movies first
        num_clips = movies_dictionary[session_key].train.shape[1] // clip_length

        if val_clip_indices is not None:
            val_clip_idx = val_clip_indices
        else:
            # Draw validation clips based on the random seed
            rnd = np.random.RandomState(seed)
            val_clip_idx = list(rnd.choice(num_clips, num_val_clips, replace=False))

        all_movies = get_movie_splits(
            movies_dictionary[session_key].train,
            movies_dictionary[session_key].test_dict,
            val_clip_idx=val_clip_idx,
            num_clips=num_clips,
            clip_length=clip_length,
        )

        # Extract all splits from neural data
        neuron_data = NeuronDataSplit(
            responses=session_data,
            val_clip_idx=val_clip_idx,
            num_clips=num_clips,
            clip_length=clip_length,
        )

        clip_chunk_sizes = {
            "train": train_chunk_size,
            "validation": clip_length,
        }
        # Create dataloaders for each fold
        for fold in ["train", "validation"]:
            dataloaders[fold][session_key] = get_movie_dataloader(
                movie=all_movies[fold],
                responses=neuron_data.response_dict[fold],
                split=fold,
                chunk_size=clip_chunk_sizes[fold],
                batch_size=batch_size,
                scene_length=clip_length,
                allow_over_boundaries=allow_over_boundaries,
            )
        # test movies
        for name in all_movies["test_dict"].keys():
            movie = all_movies["test_dict"][name]
            dataloaders[name][session_key] = get_movie_dataloader(
                movie=movie,
                responses=neuron_data.response_dict_test[name],
                split="test",
                chunk_size=movie.shape[1],
                batch_size=batch_size,
                scene_length=clip_length,
                allow_over_boundaries=allow_over_boundaries,
            )

    return dataloaders



import argparse
import os
from functools import partial
from pathlib import PurePath
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch

from openretina.data_io.hoefling_2024.constants import STIMULUS_RANGE_CONSTRAINTS
from openretina.insilico.stimulus_optimization.objective import (
    IncreaseObjective,
    InnerNeuronVisualizationObjective,
    SliceMeanReducer,
)
from openretina.insilico.stimulus_optimization.optimization_stopper import OptimizationStopper
from openretina.insilico.stimulus_optimization.optimizer import optimize_stimulus
from openretina.insilico.stimulus_optimization.regularizer import (
    ChangeNormJointlyClipRangeSeparately,
    RangeRegularizationLoss,
)
from openretina.models.core_readout import load_core_readout_model
from openretina.utils.nnfabrik_model_loading import Center, load_ensemble_model_from_remote
from openretina.utils.plotting import plot_stimulus_composition, save_stimulus_to_mp4_video


def add_parser_arguments(parser: argparse.ArgumentParser):
    parser.description = (
        "Visualize most exciting stimuli for each neuron in the model and visualize the weights of the model."
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--model-path", type=str)
    group.add_argument("--is-hoefling-ensemble-model", action="store_true")

    parser.add_argument(
        "--save-folder",
        type=str,
        required=True,
        help="Path were to save outputs",
    )
    parser.add_argument(
        "--device", type=str, choices=["cuda", "cpu"], default="cuda" if torch.cuda.is_available() else "cpu"
    )
    parser.add_argument(
        "--model-id",
        type=int,
        default=-1,
        help="If >= 0 load the ensemble model with that model_id, else use torch.load to load the model",
    )
    parser.add_argument(
        "--time-steps-stimulus",
        type=int,
        default=50,
        help="The time steps used in the stimulus to optimize",
    )
    parser.add_argument(
        "--image-file-format", type=str, default="jpg", help="File format to save the visualization plots in"
    )


def load_model(
    path: str | None,
    device: str = "cuda",
    model_id: int = 0,
    do_center_readout: bool = False,
    is_hoefling_ensemble_model: bool = False,
):
    if path is not None:
        # check if the filename contains gru to decide whether to load from the gru name
        is_gru_model = "gru" in PurePath(path).name.lower()
        model = load_core_readout_model(path, device, is_gru_model)
        print(f"Initializing lightning model from {path} to {device=} ({is_gru_model=}")
    elif is_hoefling_ensemble_model:
        center_readout = Center(target_mean=(0.0, 0.0)) if do_center_readout else None
        _, ensemble_model = load_ensemble_model_from_remote(device=device, center_readout=center_readout)
        print("Initialized ensemble model from remote.")
        model = ensemble_model.members[model_id]
    else:
        raise ValueError(
            f"Either path must be set or is_hoefling_ensemble_model must be True "
            f"({path=}, {is_hoefling_ensemble_model=})"
        )

    return model


def _get_min_max_values_and_norm(num_channels: int) -> tuple[list[tuple], float | None]:
    if num_channels == 2:
        min_max_values = [
            (STIMULUS_RANGE_CONSTRAINTS["x_min_green"], STIMULUS_RANGE_CONSTRAINTS["x_max_green"]),
            (STIMULUS_RANGE_CONSTRAINTS["x_min_uv"], STIMULUS_RANGE_CONSTRAINTS["x_max_uv"]),
        ]
        norm = float(STIMULUS_RANGE_CONSTRAINTS["norm"])
        return min_max_values, norm
    else:
        return [(None, None)], None


def visualize_model_neurons(
    model_path: str,
    save_folder: str,
    device: str,
    model_id: int,
    is_hoefling_ensemble_model: bool,
    image_file_format: str,
    time_steps_stimulus: int,
) -> None:
    model = load_model(
        model_path,
        device=device,
        model_id=model_id,
        do_center_readout=True,
        is_hoefling_ensemble_model=is_hoefling_ensemble_model,
    )
    model.eval()

    stimulus_shape = model.stimulus_shape(time_steps=time_steps_stimulus, num_batches=1)
    response_reducer = SliceMeanReducer(axis=0, start=10, length=10)
    min_max_values, norm = _get_min_max_values_and_norm(stimulus_shape[1])
    stimulus_postprocessor = ChangeNormJointlyClipRangeSeparately(
        min_max_values=min_max_values,
        norm=norm,
    )
    stimulus_regularizing_loss = RangeRegularizationLoss(
        min_max_values=min_max_values,
        max_norm=norm,
        factor=0.1,
    )
    optimizer_init_fn = partial(torch.optim.SGD, lr=100.0)

    try:
        model_readout_keys = model.readout_keys()
    except AttributeError:
        model_readout_keys = model.readout.readout_keys()
    data_key = model_readout_keys[0]
    inner_neuron_objective = InnerNeuronVisualizationObjective(model, data_key, response_reducer)
    # only select output of each layer (ignore submodules like ..._layer0_norm or ..._layer0_pool)
    layer_names_array = [x for x in inner_neuron_objective.features_dict.keys() if "layer" in x and x[-1].isdigit()]
    print(f"Generating MEIs for the following layers: {layer_names_array}")
    for layer_name in layer_names_array:
        output_shape = inner_neuron_objective.get_output_shape_for_layer(layer_name, stimulus_shape)
        if output_shape is None:
            print(f"Could not determine output shape for {layer_name=}, skipping layer.")
            continue
        num_channels, num_timesteps = output_shape[1:3]
        # We maximize the last frames of the time dimension of the output of the layer
        response_reducer.start = num_timesteps - response_reducer.length
        print(f"Reset response reduce for layer {layer_name} to: {response_reducer}")
        for channel_id in range(num_channels):
            print(f"Optimizing {layer_name=} {channel_id=}")
            stimulus = torch.randn(stimulus_shape, requires_grad=True, device=device)
            stimulus.data = stimulus_postprocessor.process(stimulus.data * 0.1)

            inner_neuron_objective.set_layer_channel(layer_name, channel_id)

            try:
                optimize_stimulus(
                    stimulus,
                    optimizer_init_fn,
                    inner_neuron_objective,
                    OptimizationStopper(max_iterations=10),
                    stimulus_regularization_loss=stimulus_regularizing_loss,
                    stimulus_postprocessor=stimulus_postprocessor,
                )
            except Exception as e:
                print(f"Skipping {layer_name=} {channel_id=} because of Exception: {e}")
                continue
            fig_axes_tuple = plt.subplots(2, 2, figsize=(7 * 3, 12))
            axes: np.ndarray[Any, plt.Axes] = fig_axes_tuple[1]  # type: ignore

            highlight_stim_start = stimulus_shape[2] - num_timesteps + response_reducer.start
            highlight_stim_end = highlight_stim_start + response_reducer.length - 1
            plot_stimulus_composition(
                stimulus=stimulus[0],
                temporal_trace_ax=axes[0, 0],
                freq_ax=axes[0, 1],
                spatial_ax=axes[1, 0],
                highlight_x_list=[(highlight_stim_start, highlight_stim_end)],
            )
            output_folder = f"{save_folder}/{layer_name}"
            os.makedirs(output_folder, exist_ok=True)
            fig_path = f"{output_folder}/{channel_id}.{image_file_format}"
            fig_axes_tuple[0].savefig(fig_path, bbox_inches="tight", facecolor="w", dpi=300)
            print(f"Saved figure at {fig_path=}")
            fig_axes_tuple[0].clf()
            plt.close()
            save_stimulus_to_mp4_video(stimulus[0], f"{output_folder}/{channel_id}.mp4")

    response_reducer = SliceMeanReducer(axis=0, start=10, length=10)
    print(f"Reset response reducer for optimizing output neurons: {response_reducer}")
    for session_key in model_readout_keys:
        output_folder = f"{save_folder}/output_neurons/{session_key}"
        os.makedirs(output_folder, exist_ok=True)
        print(f"Optimizing output neurons for {session_key} in folder {output_folder}")

        for neuron_id in range(model.readout[session_key].outdims):
            objective = IncreaseObjective(
                model, neuron_indices=neuron_id, data_key=session_key, response_reducer=response_reducer
            )
            stimulus = torch.randn(stimulus_shape, requires_grad=True, device=device)
            stimulus.data = stimulus_postprocessor.process(stimulus.data * 0.1)

            try:
                optimize_stimulus(
                    stimulus,
                    optimizer_init_fn,
                    objective,
                    OptimizationStopper(max_iterations=10),
                    stimulus_regularization_loss=stimulus_regularizing_loss,
                    stimulus_postprocessor=stimulus_postprocessor,
                )
            except Exception as e:
                print(f"Skipping neuron {neuron_id} in session {session_key} because of exception {e}")
                continue
            fig_axes_tuple = plt.subplots(2, 2, figsize=(7 * 3, 12))
            axes: np.ndarray[Any, plt.Axes] = fig_axes_tuple[1]  # type: ignore

            plot_stimulus_composition(
                stimulus=stimulus[0],
                temporal_trace_ax=axes[0, 0],
                freq_ax=axes[0, 1],
                spatial_ax=axes[1, 0],
                highlight_x_list=[(40, 49)],
            )
            fig_path = f"{output_folder}/{neuron_id}.{image_file_format}"
            fig_axes_tuple[0].savefig(fig_path, bbox_inches="tight", facecolor="w", dpi=300)
            fig_axes_tuple[0].clf()
            save_stimulus_to_mp4_video(stimulus[0], f"{output_folder}/{neuron_id}.mp4")
            plt.close()

    # Reload model without centered readouts
    if not is_hoefling_ensemble_model:
        model.save_weight_visualizations(save_folder)
    else:
        model = load_model(
            model_path,
            device=device,
            model_id=model_id,
            do_center_readout=False,
            is_hoefling_ensemble_model=is_hoefling_ensemble_model,
        )
        model.to(device).eval()
        for session_key in model_readout_keys:
            folder_path = f"{save_folder}/weights_readout/{session_key}"
            os.makedirs(folder_path, exist_ok=True)
            model.readout[session_key].save_weight_visualizations(folder_path, image_file_format)
            print(f"Plotted visualizations for {session_key}")

        # Visualize weights
        for i, layer in enumerate(model.core.features):
            output_dir = f"{save_folder}/weights_layer_{i}"
            os.makedirs(output_dir, exist_ok=True)
            layer.conv.save_weight_visualizations(output_dir, image_file_format)
            print(f"Saved weight visualization at path {output_dir}")



import numpy as np
import scipy


def butter_lowpass_filter(data: np.ndarray, lowpass_cutoff: float, fs: float, order: int = 5):
    b, a = scipy.signal.butter(order, Wn=lowpass_cutoff, fs=fs, btype="low")
    y = scipy.signal.filtfilt(b, a, data)
    return y


def calculate_fft(
    temporal_kernel: np.ndarray,
    sampling_frequency: float = 30.0,
    lowpass_cutoff: float = 10.0,
) -> tuple[np.ndarray, np.ndarray]:
    y = butter_lowpass_filter(temporal_kernel, lowpass_cutoff=lowpass_cutoff, fs=sampling_frequency)
    n_samples = temporal_kernel.shape[0]
    fft_frequencies = scipy.fft.fftfreq(n_samples, 1 / sampling_frequency)
    fft_values = scipy.fft.fft(y)

    fft_frequencies = fft_frequencies[: n_samples // 2]
    fft_weights = np.abs(fft_values[: n_samples // 2])

    return fft_frequencies, fft_weights


def weighted_main_frequency(fft_frequencies: np.ndarray, fft_weights: np.ndarray) -> float:
    assert len(fft_frequencies.shape) == 1
    assert fft_frequencies.shape == fft_weights.shape
    return np.average(fft_frequencies, weights=fft_weights)


def decompose_kernel(
    space_time_kernel: np.ndarray, scaling_factor: float = 1.0
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Scale spatial and temporal components such that spatial component is in the range [-1, 1]
    Args:
        space_time_kernel: shape=[time, y_shape, x.shape]
        scaling_factor: optional scaling factor
    Returns:
    """
    assert len(space_time_kernel.shape) == 3
    dt, dy, dx = space_time_kernel.shape

    space_time_kernel_flat = space_time_kernel.reshape((dt, dy * dx))
    U, S, V = scipy.linalg.svd(space_time_kernel_flat)
    temporal = U[:, 0]
    spatial = V[0]
    scaling_factor *= S[0]
    abs_max_val = max(np.abs(spatial.min()), np.abs(spatial.max()))
    spatial = spatial / abs_max_val
    temporal = temporal * abs_max_val * scaling_factor
    reshaped_spatial = spatial.reshape((dy, dx))
    center_x, center_y = int(dy / 2), int(dx / 2)
    if np.mean(reshaped_spatial[center_x - 3 : center_x + 2, center_y - 3 : center_y + 2]) < 0:
        spatial *= -1
        temporal *= -1
    singular_values = S[1]
    return temporal, reshaped_spatial, singular_values



import warnings

import numpy as np
import torch
from jaxtyping import Float

from openretina.utils.constants import EPSILON
from openretina.utils.misc import tensors_to_device
from openretina.utils.model_utils import eval_state


def correlation_numpy(
    y1: np.ndarray, y2: np.ndarray, axis: None | int | tuple[int, ...] = -1, eps: float = 1e-8, **kwargs
) -> np.ndarray:
    """Compute the correlation between two NumPy arrays along the specified dimension(s)."""
    y1 = (y1 - y1.mean(axis=axis, keepdims=True)) / (y1.std(axis=axis, keepdims=True, ddof=0) + eps)
    y2 = (y2 - y2.mean(axis=axis, keepdims=True)) / (y2.std(axis=axis, keepdims=True, ddof=0) + eps)
    corr = (y1 * y2).mean(axis=axis, **kwargs)
    return corr


def MSE_numpy(y1: np.ndarray, y2: np.ndarray, axis: None | int | tuple[int, ...] = -1, **kwargs) -> np.ndarray:
    """Compute the mean squared error between two NumPy arrays along the specified dimension(s)."""
    return ((y1 - y2) ** 2).mean(axis=axis, **kwargs)


def poisson_loss_numpy(
    y_true: np.ndarray, y_pred: np.ndarray, eps: float = 1e-8, mean_axis: None | int | tuple[int, ...] = -1
) -> np.ndarray:
    """Compute the Poisson loss between two NumPy arrays."""
    return (y_pred - y_true * np.log(y_pred + eps)).mean(axis=mean_axis)


def model_predictions(loader, model: torch.nn.Module, data_key, device) -> tuple[np.ndarray, np.ndarray]:
    """
    computes model predictions for a given dataloader and a model
    Returns:
        target: ground truth, i.e. neuronal firing rates of the neurons
        output: responses as predicted by the network
    """
    target, output = torch.empty(0), torch.empty(0)
    for *inputs, responses in loader[data_key]:  # tuple unpacking necessary for group assignments when present
        output = torch.cat(
            (output, (model(*tensors_to_device(inputs, device), data_key=data_key).detach().cpu())), dim=0
        )
        target = torch.cat((target, responses.detach().cpu()), dim=0)
    output_np = output.numpy()
    target_np = target.numpy()
    lag = target_np.shape[1] - output_np.shape[1]

    return target_np[:, lag:, ...], output_np


def corr_stop(model: torch.nn.Module, loader, avg: bool = True, device: str = "cpu"):
    """
    Returns either the average correlation of all neurons or the correlations per neuron.
        Gets called by early stopping and the model performance evaluation
    """

    n_neurons, correlations_sum = 0, 0
    if not avg:
        all_correlations = np.array([])

    for data_key in loader:
        with eval_state(model):
            target, output = model_predictions(loader, model, data_key, device)

        ret = correlation_numpy(target, output, axis=0)

        if np.any(np.isnan(ret)):
            warnings.warn(f"{np.isnan(ret).mean() * 100}% NaNs ")
        ret[np.isnan(ret)] = 0

        if not avg:
            all_correlations = np.append(all_correlations, ret)
        else:
            n_neurons += output.shape[1]
            correlations_sum += ret.sum()

    corr_ret = correlations_sum / n_neurons if avg else all_correlations
    return corr_ret


def corr_stop3d(model: torch.nn.Module, loader, avg: bool = True, device: str = "cpu"):
    """
    Returns either the average correlation of all neurons or the correlations per neuron.
        Gets called by early stopping and the model performance evaluation
    """

    n_neurons, correlations_sum = 0, 0
    if not avg:
        all_correlations = np.array([])

    for data_key in loader:
        with eval_state(model):
            target, output = model_predictions(loader, model, data_key, device)

        # Correlation over time axis (1)
        ret = correlation_numpy(target, output, axis=1)

        # Average over batches
        ret = ret.mean(axis=0)

        if np.any(np.isnan(ret)):
            warnings.warn(f"{np.isnan(ret).mean() * 100}% NaNs in corr_stop3d")
        ret[np.isnan(ret)] = 0

        if not avg:
            all_correlations = np.append(all_correlations, ret)
        else:
            n_neurons += output.shape[-1]
            correlations_sum += ret.sum()

    corr_ret = correlations_sum / n_neurons if avg else all_correlations
    return corr_ret


def poisson_stop(model: torch.nn.Module, loader, avg: bool = False, device: str = "cpu"):
    poisson_losses = np.array([])
    n_neurons = 0
    for data_key in loader:
        with eval_state(model):
            target, output = model_predictions(loader, model, data_key, device)

        ret = output - target * np.log(output + EPSILON)
        if np.any(np.isnan(ret)):
            warnings.warn(f" {np.isnan(ret).mean() * 100}% NaNs ")

        poisson_losses = np.append(poisson_losses, np.nansum(ret, 0))
        n_neurons += output.shape[1]
    return poisson_losses.sum() / n_neurons if avg else poisson_losses.sum()


def poisson_stop3d(model: torch.nn.Module, loader, avg: bool = True, device: str = "cpu"):
    poisson_losses = np.array([])
    n_neurons = 0
    n_batch = 0
    for data_key in loader:
        with eval_state(model):
            target, output = model_predictions(loader, model, data_key, device)

        ret = output - target * np.log(output + EPSILON)
        if np.any(np.isnan(ret)):
            warnings.warn(f"{np.isnan(ret).mean() * 100:.2f} % NaNs in poisson_stop3d")

        poisson_losses = np.append(poisson_losses, np.nansum(ret, axis=(0, 1)))  # sum over batches (0) and time (1)
        n_neurons += output.shape[-1]
        n_batch += output.shape[0]
    return poisson_losses.sum() / (n_neurons * n_batch) if avg else poisson_losses.sum()


def MSE_stop3d(model: torch.nn.Module, loader, avg: bool = True, device: str = "cpu"):
    mse_losses = np.array([])
    n_neurons = 0
    n_batch = 0
    for data_key in loader:
        with eval_state(model):
            target, output = model_predictions(loader, model, data_key, device)

        ret = (output - target) ** 2
        if np.any(np.isnan(ret)):
            warnings.warn(f"{np.isnan(ret).mean:.1%} NaNs")

        mse_losses = np.append(mse_losses, np.nansum(ret, axis=(0, 1)))  # sum over batches (0) and time (1)
        n_neurons += output.shape[-1]
        n_batch += output.shape[0]
    return mse_losses.sum() / (n_neurons * n_batch) if avg else mse_losses.sum()


def explainable_vs_total_var(
    repeated_outputs: Float[np.ndarray, "frames repeats neurons"], eps: float = 1e-9
) -> tuple[Float[np.ndarray, " neurons"], Float[np.ndarray, " neurons"]]:
    """
    Adapted from neuralpredictors.
    Compute the ratio of explainable to total variance per neuron.
    See Cadena et al., 2019: https://doi.org/10.1371/journal.pcbi.1006897

    Args:
        repeated_outputs (array): numpy array with shape (images/time, repeats, neurons) containing the responses.

    Returns:
        tuple: A tuple containing:
            - var_ratio (array): Ratio of explainable to total variance per neuron
            - explainable_var (array): Explainable variance for each neuron
    """
    total_var = np.var(repeated_outputs, axis=(0, 1), ddof=1)
    repeats_var = np.var(repeated_outputs, axis=1, ddof=1)
    noise_var = np.mean(repeats_var, axis=0)
    # Clip. In some bad cases, noise_var can be larger than total_var.
    explainable_var = np.clip(total_var - noise_var, eps, None)
    var_ratio = explainable_var / (total_var + eps)
    return var_ratio, explainable_var


def feve(
    targets: Float[np.ndarray, "frames repeats neurons"],
    predictions: Float[np.ndarray, "frames repeats neurons"] | Float[np.ndarray, "frames neurons"],
) -> Float[np.ndarray, " neurons"]:
    """
    Adapted from neuralpredictors.
    Compute the fraction of explainable variance explained per neuron

    Args:
        targets (array-like): Neuron responses (ground truth) over time / different images across repetitions.
        Dimensions: np.array(images/time, num_repeats, num_neurons)
        predictions (array-like): Model predictions to the repeated images, either including or excluding
        repetitions. Dimensions: np.array(images/time, num_repeats, num_neurons) or np.array(images/time, num_neurons)
    Returns:
        FEVe (np.array): the fraction of explainable variance explained per neuron

    """
    if predictions.shape[1] != targets.shape[1] and predictions.ndim == 2:
        predictions = np.repeat(predictions[:, np.newaxis, :], targets.shape[1], axis=1)

    assert targets.shape == predictions.shape, (
        f"Targets and predictions must have the same shape, got {targets.shape} and {predictions.shape}"
    )

    sum_square_res = [(target - prediction) ** 2 for target, prediction in zip(targets, predictions)]
    sum_square_res = np.concatenate(sum_square_res, axis=0)

    var_ratio, explainable_var = explainable_vs_total_var(targets)
    # Invert the formula to get the noise variance
    total_var = explainable_var / var_ratio
    noise_var = total_var - explainable_var

    mse = np.mean(sum_square_res, axis=0)  # mean over time and reps
    fev_e = 1 - np.clip(mse - noise_var, 0, None) / (explainable_var)
    return np.clip(fev_e, 0, None)


def crop_responses(responses: np.ndarray, predictions: np.ndarray) -> np.ndarray:
    """

    :param responses: array of responses, last axis is time
    :param predictions: array of predictions, first axis is time
    :return: responses cropped to same shape as predictions
    """

    lag = responses.shape[-1] - predictions.shape[0]
    return responses[..., lag:]



"""
Minimal responses loading utilities to train a model on the data used in Maheswaranathan et al. 2023

Paper: https://doi.org/10.1016/j.neuron.2023.06.007
Data: https://doi.org/10.25740/rk663dm5577

OpenRetina provides a mirror of the dataset on huggingface:
https://huggingface.co/datasets/open-retina/open-retina/resolve/main/baccus_lab/maheswaranathan_2023/neural_code_data.zip
"""

import os
from typing import Literal

from openretina.data_io.base import ResponsesTrainTestSplit
from openretina.data_io.base_dataloader import NeuronDataSplit
from openretina.utils.file_utils import get_local_file_path
from openretina.utils.h5_handling import load_dataset_from_h5


def load_all_responses(
    base_data_path: str | os.PathLike,
    response_type: Literal["firing_rate_5ms", "firing_rate_10ms", "firing_rate_20ms"] = "firing_rate_20ms",
    stim_type: Literal["naturalscene", "whitenoise"] = "naturalscene",
    fr_normalization: float = 1,
) -> dict[str, ResponsesTrainTestSplit]:
    """
    Load all neural responses from sessions within subfolders in a given base data path.

    The base data path should point to the location of of the `neural_code_data` folder.
    (See https://doi.org/10.25740/rk663dm5577 for dataset download).

    Alternatively, base_data_path can point directly to our huggingface mirror of the dataset, which will then
    be downloaded and extracted automatically to the openretina cache directory.
    https://huggingface.co/datasets/open-retina/open-retina/resolve/main/baccus_lab/maheswaranathan_2023/neural_code_data.zip
    """
    # Resolve data path
    base_data_path = get_local_file_path(str(base_data_path))
    full_data_path = os.path.join(base_data_path, "ganglion_cell_data")

    responses_all_sessions = {}
    for session in [x.name for x in os.scandir(os.fspath(full_data_path)) if x.is_dir()]:
        session_path = os.path.join(full_data_path, session)
        for recording_file in [x for x in os.listdir(session_path) if str(x).endswith(f"{stim_type}.h5")]:
            recording_file = os.path.join(session_path, recording_file)

            print(f"Loading responses from {recording_file}")

            # Load neural responses
            train_session_data = load_dataset_from_h5(recording_file, f"/train/response/{response_type}")
            test_session_data = load_dataset_from_h5(recording_file, f"/test/response/{response_type}")

            assert train_session_data.shape[0] == test_session_data.shape[0], (
                "Train and test responses should have the same number of neurons."
            )

            responses_all_sessions[str(session)] = ResponsesTrainTestSplit(
                train=train_session_data / fr_normalization,
                test=test_session_data / fr_normalization,
                stim_id=stim_type,
            )
    return responses_all_sessions


class NeuronDataBaccus(NeuronDataSplit):
    pass



import argparse
import os
import sys
from pathlib import Path

import hydra
from omegaconf import DictConfig

from openretina.cli import create_data, visualize_model_neurons
from openretina.utils.file_utils import get_cache_directory


def get_config_path(config_path: str | None) -> str:
    """Get the absolute path to the configs directory"""
    if config_path is None:
        path = str(Path(__file__).parent.parent.parent / "configs")
        if not os.path.isdir(path):
            raise ValueError(
                f"Config directory not found at {path}. Please specify a config directory using '--config-path'"
            )
    else:
        path = str(Path(config_path).absolute())
    return path


class HydraRunner:
    """Wrapper class to run Hydra commands"""

    @staticmethod
    def train(config_path: str, args: list[str]) -> None:
        # Modify sys.argv to work with Hydra
        sys.argv = [sys.argv[0]] + args

        @hydra.main(
            version_base="1.3",
            config_path=get_config_path(config_path),
            config_name="hoefling_2024_core_readout_high_res",
        )
        def _train(cfg: DictConfig) -> None:
            from openretina.cli.train import train_model  # Import actual training function

            # Check if cache_dir is set or left as none, in which case we set it to the cache directory
            if cfg.paths.cache_dir is None:
                cfg.paths.cache_dir = get_cache_directory()

            train_model(cfg)

        _train()


def main() -> None:
    parser = argparse.ArgumentParser(description="OpenRetina CLI")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Train command
    train_parser = subparsers.add_parser("train", help="Train a model")
    train_parser.add_argument("--config-path", default=None, help="Path to config directory")

    # Visualize command
    visualize_parser = subparsers.add_parser("visualize", help="Visualize a model")
    visualize_model_neurons.add_parser_arguments(visualize_parser)

    # Create data command
    create_data_parser = subparsers.add_parser("create-data", help="Create artificial data")
    create_data.add_parser_arguments(create_data_parser)

    args, unknown_args = parser.parse_known_args()  # this is an 'internal' method
    command = args.__dict__.pop("command")

    if command == "train":
        HydraRunner.train(args.config_path, unknown_args)
    elif command == "visualize":
        if len(unknown_args) > 0:
            print(f"Warn: found the following unknown args: {unknown_args}")
        visualize_model_neurons.visualize_model_neurons(**vars(args))
    elif command == "create-data":
        create_data.write_data_to_directory(**vars(args))
    elif command is None:
        parser.print_help()
    else:
        raise ValueError(f"Unknown command: {args.command}")



from functools import partial
import torch; import matplotlib.pyplot as plt
from openretina.models import *
from openretina.insilico import *
from openretina.utils import *

model = load_core_readout_from_remote("hoefling_2024_base_low_res", "cpu")
# Define objective that increases the activity of a single neuron between timestep 10 and 20.
objective = IncreaseObjective(
model,
neuron_indices=[10],
data_key="session_2_ventral2_20201016",
response_reducer=SliceMeanReducer(axis=0, start=10, length=10),
)
# Initialize random stimulus with a variance of 0.1
stimulus = torch.randn(model.stimulus_shape(time_steps=40), requires_grad=True)
stimulus.data = stimulus.data * 0.1
# Optimize stimulus with gradient descent towards the objective.
optimize_stimulus(
stimulus=stimulus,
optimizer_init_fn=partial(torch.optim.SGD, lr=100.0),
objective_object=objective,
optimization_stopper=OptimizationStopper(max_iterations=10),
)
# Visualize stimulus
save_stimulus_to_mp4_video(stimulus[0], "mei.mp4")
fig, axes = plt.subplots(2, 2, figsize=(7 * 3, 12))
plot_stimulus_composition(stimulus[0], axes[0,0], axes[0,1], axes[1,0])




import os.path
import tempfile

import numpy as np
import pytest

from openretina.utils.plotting import numpy_to_mp4_video, save_stimulus_to_mp4_video


@pytest.mark.parametrize(
    "stimulus_shape",
    [
        (1, 5, 7, 11),
        (2, 5, 7, 11),
        (3, 5, 7, 11),
    ],
)
def test_save_stimulus_to_mp4_video(stimulus_shape: tuple[int, ...]) -> None:
    stimulus = np.random.rand(*stimulus_shape)
    with tempfile.NamedTemporaryFile(suffix=".mp4", delete=True) as temp_file:
        save_stimulus_to_mp4_video(stimulus, filepath=temp_file.name)
        assert os.path.exists(temp_file.name)


@pytest.mark.parametrize(
    "stimulus_shape",
    [
        (11, 7, 5, 3),
    ],
)
def test_numpy_to_mp4_video(stimulus_shape: tuple[int, ...]) -> None:
    video = np.random.rand(*stimulus_shape)
    with tempfile.NamedTemporaryFile(suffix=".mp4", delete=True) as temp_file:
        numpy_to_mp4_video(video, temp_file.name, display_video=False)
        assert os.path.exists(temp_file.name)



import os
import pickle
import warnings
from dataclasses import InitVar, dataclass, field
from functools import cached_property
from typing import Any, Optional

import numpy as np
import torch
from jaxtyping import Float


@dataclass(frozen=True)
class MoviesTrainTestSplit:
    train: Float[np.ndarray, "channels train_time height width"]
    # test_dict: dict[str, Float[np.ndarray, "channels train_time height width"]]
    # (dataclass complains about unsupported value type when using full annotation)
    test_dict: dict = field(default_factory=lambda: {})
    test: InitVar[Float[np.ndarray, "channels test_time height width"] | None] = None
    stim_id: Optional[str] = None
    random_sequences: Optional[np.ndarray] = None
    norm_mean: Optional[float] = None
    norm_std: Optional[float] = None

    def __post_init__(self, test: Float[np.ndarray, "channels test_time height width"] | None):
        if (len(self.test_dict) == 0) == (test is None):
            raise ValueError(f"Exactly one of test_dict and test should be set, but {test=} {self.test_dict=}.")
        if len(self.test_dict) == 0:
            self.test_dict["test"] = test

        assert self.train.ndim == 4, "Train movie should have 4 dimensions."
        assert len(self.test_shape) == 4, "Test movie should have 4 dimensions."
        assert self.train.shape[0] == self.test_shape[0], "Channel dimension does not match in train and test movies."
        assert self.train.shape[2:] == self.test_shape[2:], "Spatial dimensions do not match in train and test movies."
        if self.train.shape[0] > self.train.shape[1]:
            warnings.warn(
                "The number of channels is greater than the number of timebins in the train movie. "
                "Check if the provided data is in the correct format.",
                category=UserWarning,
                stacklevel=2,
            )

    @cached_property
    def test_shape(self) -> tuple[int, int, int, int]:
        for n, t in self.test_dict.items():
            if t.ndim != 4:
                raise ValueError(f"Test stimulus {n} is not 4 dimensional: {t.shape}")
            max_temp_dim = max(t.shape[1] for t in self.test_dict.values())
            test_shapes = {(t.shape[0], max_temp_dim, t.shape[2], t.shape[3]) for t in self.test_dict.values()}
            if len(test_shapes) > 1:
                raise ValueError(f"Inconsistent test shapes: {test_shapes=}")
            return next(iter(test_shapes))
        raise ValueError("No test stimuli present.")

    @property
    def test_movie(self) -> np.ndarray:
        if len(self.test_dict) > 1:
            raise ValueError(f"Multiple test responses present: {list(self.test_dict.keys())}")
        return self.test_dict[next(iter(self.test_dict.keys()))]

    @classmethod
    def from_pickle(cls, file_path: str | os.PathLike):
        with open(file_path, "rb") as f:
            movies_dict = pickle.load(f)
        return cls(
            train=movies_dict["train"],
            test=movies_dict["test"],
            random_sequences=movies_dict.get("random_sequences", None),
            norm_mean=movies_dict.get("movie_stats", {}).get("dichromatic", {}).get("mean", np.array([np.nan])).item(),
            norm_std=movies_dict.get("movie_stats", {}).get("dichromatic", {}).get("sd", np.array([np.nan])).item(),
        )


@dataclass
class ResponsesTrainTestSplit:
    train: Float[np.ndarray, "neurons train_time"]
    # test_dict: dict[str, Float[np.ndarray, "neurons test_time"]]
    # (dataclass complains about unsupported value type when using full annotation)
    test_dict: dict = field(default_factory=lambda: {})
    test: InitVar[Float[np.ndarray, "neurons test_time"] | None] = None
    test_by_trial: Float[np.ndarray, "trials neurons test_time"] | None = None
    stim_id: str | None = None
    session_kwargs: dict[str, Any] = field(default_factory=lambda: {})

    def __post_init__(self, test):
        if (len(self.test_dict) == 0) == (test is None):
            raise ValueError(f"Exactly one of test_dict and test should be set, but {test=} {self.test_dict=}.")
        if len(self.test_dict) == 0:
            self.test_dict["test"] = test

        assert self.train.shape[0] == self.test_neurons, (
            "Train and test responses should have the same number of neurons."
        )
        if self.train.shape[0] > self.train.shape[1]:
            warnings.warn(
                "The number of neurons is greater than the number of timebins in the train responses. "
                "Check if the provided data is in the correct format.",
                category=UserWarning,
                stacklevel=2,
            )

    @cached_property
    def test_neurons(self) -> int:
        for name, t in self.test_dict.items():
            if t.ndim != 2:
                raise ValueError(f"Test responses for {name=} are not two dimensions: {t.shape}")
        test_neurons = set(x.shape[0] for x in self.test_dict.values())
        if len(test_neurons) > 1:
            raise ValueError(f"Test responses have inconsistent number of neurons: {test_neurons=}")
        return next(iter(test_neurons))

    def check_matching_stimulus(self, movies: MoviesTrainTestSplit):
        assert self.train.shape[1] == movies.train.shape[1], (
            "Train movie and responses should have the same timebins."
            f"Got {self.train.shape[1]} and {movies.train.shape[1]}."
        )
        assert set(self.test_dict.keys()) == set(movies.test_dict.keys()), "Test movie and responses should match"
        for k in self.test_dict.keys():
            responses = self.test_dict[k]
            movie = movies.test_dict[k]
            assert responses.shape[1] == movie.shape[1], (
                "Test movie and responses should have the same timebins.",
                f"Got {movie.shape[1]=} and {responses.shape[1]}.",
            )
        if self.stim_id is not None and movies.stim_id is not None:
            assert self.stim_id == movies.stim_id, "Stimulus ID in responses and movies do not match."

    @property
    def n_neurons(self) -> int:
        return self.train.shape[0]

    @property
    def test_response(self) -> Float[np.ndarray, "neurons test_time"]:
        if len(self.test_dict) > 1:
            raise ValueError(f"Multiple test stimuli: {list(self.test_dict.keys())}")
        return self.test_dict[next(iter(self.test_dict.keys()))]


def get_n_neurons_per_session(responses_dict: dict[str, ResponsesTrainTestSplit]) -> dict[str, int]:
    return {name: responses.n_neurons for name, responses in responses_dict.items()}


def normalize_train_test_movies(
    train: Float[np.ndarray, "channels train_time height width"],
    test: Float[np.ndarray, "channels test_time height width"],
) -> tuple[
    Float[np.ndarray, "channels train_time height width"],
    Float[np.ndarray, "channels test_time height width"],
    dict[str, float | None],
]:
    """
    z-score normalization of train and test movies using the mean and standard deviation of the train movie.

    Parameters:
    - train: train movie with shape (channels, time, height, width)
    - test: test movie with shape (channels, time, height, width)

    Returns:
    - train_video_preproc: normalized train movie
    - test_video_preproc: normalized test movie
    - norm_stats: dictionary containing the mean and standard deviation of the train movie

    Note: The functions casts the input to torch tensors to calculate the mean and standard deviation of large
    inputs more efficiently.
    """
    train_tensor = torch.tensor(train, dtype=torch.float32)
    test_tensor = torch.tensor(test, dtype=torch.float32)
    train_mean = train_tensor.mean()
    train_std = train_tensor.std()
    train_video_preproc = (train_tensor - train_mean) / train_std
    test_video = (test_tensor - train_mean) / train_std
    return (
        train_video_preproc.cpu().detach().numpy(),
        test_video.cpu().detach().numpy(),
        {"norm_mean": train_mean.item(), "norm_std": train_std.item()},
    )


def compute_data_info(
    neuron_data_dictionary: dict[str, ResponsesTrainTestSplit],
    movies_dictionary: dict[str, MoviesTrainTestSplit] | MoviesTrainTestSplit,
) -> dict[str, Any]:
    """
    Computes information related to the data used to train a model, including the number of neurons, the shape of the
    movies, and the normalization statistics. This information can be fed to and saved with the models.

    Parameters:
    - neuron_data_dictionary: dictionary of responses for each session
    - movies_dictionary: dictionary of movies for each session

    Returns:
    - data_info: dictionary containing the number of neurons, the shape of the movies, the movie normalization
        statistics, and any extra session kwargs related to the data.
    """
    n_neurons_dict = get_n_neurons_per_session(neuron_data_dictionary)
    if isinstance(movies_dictionary, MoviesTrainTestSplit):
        movies_dictionary = {"default": movies_dictionary}
    input_shape = tuple((movie.train.shape[0], *movie.train.shape[2:]) for movie in movies_dictionary.values())[0]
    sessions_kwargs = {
        session_name: responses.session_kwargs for session_name, responses in neuron_data_dictionary.items()
    }

    return {
        "n_neurons_dict": n_neurons_dict,
        "input_shape": input_shape,
        "sessions_kwargs": sessions_kwargs,
        "movie_norm_dict": {
            name: {"norm_mean": movie.norm_mean, "norm_std": movie.norm_std}
            for name, movie in movies_dictionary.items()
        },
    }



import torch


class PoissonLoss3d(torch.nn.Module):
    def __init__(self, bias: float = 1e-16, per_neuron: bool = False, avg: bool = False):
        super().__init__()
        self.bias = bias
        self.per_neuron = per_neuron
        self.avg = avg

    def forward(self, output, target):
        lag = target.size(1) - output.size(1)
        loss = output - target[:, lag:, :] * torch.log(output + self.bias)
        if not self.per_neuron:
            return loss.mean() if self.avg else loss.sum()  # loss.sum(axis=(-2)).mean()
        else:
            return loss.view(-1, loss.shape[-1]).mean(dim=0)

    def __str__(self):
        bias, per_neuron, avg = self.bias, self.per_neuron, self.avg
        return f"PoissonLoss3d({bias=} {per_neuron=} {avg=})"


class CelltypePoissonLoss3d(torch.nn.Module):
    def __init__(self, bias: float = 1e-16, per_neuron: bool = False, avg: bool = False):
        super().__init__()
        self.bias = bias
        self.per_neuron = per_neuron
        self.avg = avg

    def forward(self, output, target, group_assignment, group_counts):
        lag = target.size(1) - output.size(1)
        loss = output - target[:, lag:, :] * torch.log(output + self.bias)
        loss = loss * (group_counts.sum() / group_counts.size(0) / group_counts[group_assignment - 1])[:, None, ...]
        if not self.per_neuron:
            return loss.mean() if self.avg else loss.sum()  # loss.sum(axis=(-2)).mean()
        else:
            return loss.view(-1, loss.shape[-1]).mean(dim=0)



from builtins import int
from collections import OrderedDict
from typing import Any

import torch
import torch.nn as nn

from openretina.modules.core.base_core import Core3d
from openretina.modules.layers import Bias3DLayer, Scale2DLayer, Scale3DLayer
from openretina.modules.layers.convolutions import (
    STSeparableBatchConv3d,
    TimeIndependentConv3D,
    TorchFullConv3D,
    TorchSTSeparableConv3D,
    compute_temporal_kernel,
    temporal_smoothing,
)
from openretina.modules.layers.gru import GRU_Module
from openretina.modules.layers.regularizers import FlatLaplaceL23dnorm, TimeLaplaceL23dnorm
from openretina.modules.layers.scaling import FiLM


class ConvGRUCore(Core3d, nn.Module):
    def __init__(
        self,
        n_neurons_dict: dict[str, int] | None = None,
        input_channels: int = 2,
        hidden_channels=(8,),
        temporal_kernel_size=(21,),
        spatial_kernel_size=(14,),
        layers: int = 1,
        gamma_hidden: float = 0.0,
        gamma_input: float = 0.0,
        gamma_in_sparse: float = 0.0,
        gamma_temporal: float = 0.0,
        final_nonlinearity: bool = True,
        bias: bool = True,
        input_padding: bool = False,
        hidden_padding: bool = True,
        batch_norm: bool = True,
        batch_norm_scale: bool = True,
        batch_norm_momentum: float = 0.1,
        laplace_padding: int | None = 0,
        batch_adaptation: bool = False,
        use_avg_reg: bool = False,
        nonlinearity: str = "ELU",
        conv_type: str = "custom_separable",
        use_gru: bool = False,
        use_projections: bool = False,
        gru_kwargs: dict[str, int | float] | None = None,
        **kwargs,
    ):
        super().__init__()
        # Set regularizers
        self._input_weights_regularizer_spatial = FlatLaplaceL23dnorm(padding=laplace_padding)
        self._input_weights_regularizer_temporal = TimeLaplaceL23dnorm(padding=laplace_padding)

        # Get convolution class
        self.conv_class = self.get_conv_class(conv_type)

        if n_neurons_dict is None:
            n_neurons_dict = {}
            if batch_adaptation:
                raise ValueError(
                    "If batch_adaptation is True, n_neurons_dict must be provided to "
                    "learn the adaptation terms per session."
                )

        self.layers = layers
        self.gamma_input = gamma_input
        self.gamma_in_sparse = gamma_in_sparse
        self.gamma_hidden = gamma_hidden
        self.gamma_temporal = gamma_temporal
        self.input_channels = input_channels
        self.hidden_channels = hidden_channels
        self.use_avg_reg = use_avg_reg

        if not isinstance(hidden_channels, (list, tuple)):
            hidden_channels = [hidden_channels] * (self.layers)
        if not isinstance(temporal_kernel_size, (list, tuple)):
            temporal_kernel_size = [temporal_kernel_size] * (self.layers)
        if not isinstance(spatial_kernel_size, (list, tuple)):
            spatial_kernel_size = [spatial_kernel_size] * (self.layers)

        self.features = nn.Sequential()

        # Log speed dictionary
        log_speed_dict = self.generate_log_speed_dict(n_neurons_dict, batch_adaptation) if batch_adaptation else {}

        # Padding logic
        self.input_pad, self.hidden_pad = self.calculate_padding(input_padding, hidden_padding, spatial_kernel_size)

        # Initialize layers, including projection if applicable
        self.initialize_layers(
            input_channels,
            hidden_channels,
            temporal_kernel_size,
            spatial_kernel_size,
            log_speed_dict,
            batch_norm,
            batch_norm_momentum,
            bias,
            batch_norm_scale,
            final_nonlinearity,
            self.input_pad,
            self.hidden_pad,
            nonlinearity,
            use_projections,
        )

        self.apply(self.init_conv)

        # GRU integration
        if use_gru:
            print("Using GRU")
            self.features.add_module("gru", GRU_Module(input_channels=hidden_channels[-1], **gru_kwargs))  # type: ignore

    def forward(self, input_, data_key=None):
        ret = []
        do_skip = False
        for layer_num, feat in enumerate(self.features):
            input_ = feat(
                (
                    torch.cat(ret[-min(self.skip, layer_num) :], dim=1) if do_skip else input_,
                    data_key,
                )
            )

        return input_

    def get_conv_class(self, conv_type: str) -> type[nn.Module]:
        if conv_type == "separable":
            return TorchSTSeparableConv3D
        elif conv_type == "custom_separable":
            return STSeparableBatchConv3d
        elif conv_type == "full":
            return TorchFullConv3D
        elif conv_type == "time_independent":
            return TimeIndependentConv3D
        else:
            raise ValueError(f"Un-implemented conv_type {conv_type}")

    def calculate_padding(self, input_padding, hidden_padding, spatial_kernel_size):
        if input_padding:
            input_pad = (0, spatial_kernel_size[0] // 2, spatial_kernel_size[0] // 2)
        else:
            input_pad = 0

        hidden_pad = [
            (0, spatial_kernel_size[x] // 2, spatial_kernel_size[x] // 2) if hidden_padding and x > 0 else 0
            for x in range(len(spatial_kernel_size))
        ]
        return input_pad, hidden_pad

    def initialize_layers(
        self,
        input_channels,
        hidden_channels,
        temporal_kernel_size,
        spatial_kernel_size,
        log_speed_dict,
        batch_norm,
        batch_norm_momentum,
        bias: bool,
        batch_norm_scale,
        final_nonlinearity,
        input_pad,
        hidden_pad,
        nonlinearity,
        use_projections: bool,
    ):
        layer: OrderedDict[str, Any] = OrderedDict()
        layer["conv"] = self.conv_class(
            input_channels,
            hidden_channels[0],
            log_speed_dict,
            temporal_kernel_size[0],
            spatial_kernel_size[0],
            bias=False,
            padding=input_pad,  # type: ignore
        )
        if batch_norm:
            layer["norm"] = nn.BatchNorm3d(
                hidden_channels[0],  # type: ignore
                momentum=batch_norm_momentum,
                affine=bias and batch_norm_scale,
            )  # ok or should we ensure same batch norm?
            if bias:
                if not batch_norm_scale:
                    layer["bias"] = Bias3DLayer(hidden_channels[0])
            elif batch_norm_scale:
                layer["scale"] = Scale3DLayer(hidden_channels[0])
        if final_nonlinearity:
            layer["nonlin"] = getattr(nn, nonlinearity)()
        self.features.add_module("layer0", nn.Sequential(layer))

        if use_projections:
            self.features.add_module(
                "projection",
                nn.Sequential(
                    TorchFullConv3D(
                        hidden_channels[0],
                        hidden_channels[0],
                        log_speed_dict,
                        1,
                        1,
                        bias=False,
                    ),
                    getattr(nn, nonlinearity)(),
                ),
            )

        # --- other layers

        for layer_num in range(1, self.layers):
            layer = OrderedDict()
            layer["conv"] = self.conv_class(
                hidden_channels[layer_num - 1],
                hidden_channels[layer_num],
                log_speed_dict,
                temporal_kernel_size[layer_num],
                spatial_kernel_size[layer_num],
                bias=False,
                padding=hidden_pad[layer_num - 1],  # type: ignore
            )
            if batch_norm:
                layer["norm"] = nn.BatchNorm3d(
                    hidden_channels[layer_num],
                    momentum=batch_norm_momentum,
                    affine=bias and batch_norm_scale,
                )
                if bias:
                    if not batch_norm_scale:
                        layer["bias"] = Bias3DLayer(hidden_channels[layer_num])
                elif batch_norm_scale:
                    layer["scale"] = Scale2DLayer(hidden_channels[layer_num])
            if final_nonlinearity or layer_num < self.layers - 1:
                layer["nonlin"] = getattr(nn, nonlinearity)()
            self.features.add_module(f"layer{layer_num}", nn.Sequential(layer))

    def generate_log_speed_dict(self, n_neurons_dict, batch_adaptation):
        log_speed_dict = {}
        for k in n_neurons_dict:
            var_name = "_".join(["log_speed", k])
            log_speed_val = torch.nn.Parameter(data=torch.zeros(1), requires_grad=batch_adaptation)
            setattr(self, var_name, log_speed_val)
            log_speed_dict[var_name] = log_speed_val
        return log_speed_dict

    def spatial_laplace(self):
        return self._input_weights_regularizer_spatial(self.features[0].conv.weight_spatial, avg=self.use_avg_reg)

    def group_sparsity(self):  # check if this is really what we want
        sparsity_loss = 0
        for layer in self.features:
            if hasattr(layer, "conv"):
                spatial_weight_layer = layer.conv.weight_spatial
                norm = spatial_weight_layer.pow(2).sum([2, 3, 4]).sqrt().sum(1)
                sparsity_loss_layer = (spatial_weight_layer.pow(2).sum([2, 3, 4]).sqrt().sum(1) / norm).sum()
                sparsity_loss += sparsity_loss_layer
            else:
                continue
        return sparsity_loss

    def group_sparsity0(self):  # check if this is really what we want
        weight_temporal = compute_temporal_kernel(
            torch.zeros(1, device=self.features[0].conv.sin_weights.device),
            self.features[0].conv.sin_weights,
            self.features[0].conv.cos_weights,
            self.features[0].conv.temporal_kernel_size,
        )
        # abc are dummy dimensions
        weight = torch.einsum("oitab,oichw->oithw", weight_temporal, self.features[0].conv.weight_spatial)
        return (weight.pow(2).sum([2, 3, 4]).sqrt().sum(1) / torch.sqrt(1e-8 + weight.pow(2).sum([1, 2, 3, 4]))).sum()

    def temporal_smoothness(self):
        return sum(
            temporal_smoothing(
                layer.conv.sin_weights,
                layer.conv.cos_weights,
            )
            for layer in self.features
            if hasattr(layer, "conv")
        )

    def regularizer(self):
        if self.conv_class == STSeparableBatchConv3d:
            return (
                self.group_sparsity() * self.gamma_hidden
                + self.gamma_input * self.spatial_laplace()
                + self.gamma_temporal * self.temporal_smoothness()
                + self.group_sparsity0() * self.gamma_in_sparse
            )
        else:
            return 0

    @property
    def outchannels(self):
        return len(self.features) * self.hidden_channels[-1]


class ConditionedGRUCore(ConvGRUCore, nn.Module):
    def __init__(self, cond_dim, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cond_dim = cond_dim

        j = 0
        for i, layer in enumerate(self.features):
            if hasattr(layer, "conv") and hasattr(layer, "norm"):
                setattr(self, f"film_{i}", FiLM(self.hidden_channels[j], self.cond_dim))
                # Manual instead of enumerate because of projection layers.
                j += 1

    def forward(self, x, conditioning=None, data_key=None):
        if conditioning is None:
            conditioning = torch.zeros(x.size(0), self.cond_dim, device=x.device)
        for layer_num, feat in enumerate(self.features):
            x = feat(
                (
                    x,
                    data_key,
                )
            )
            if hasattr(feat, "conv") and hasattr(feat, "norm"):
                x = getattr(self, f"film_{layer_num}")(x, conditioning)

        return x



import os
import warnings
from collections import OrderedDict

import torch
import torch.nn as nn
from matplotlib import pyplot as plt

from openretina.modules.layers import FlatLaplaceL23dnorm
from openretina.modules.layers.convolutions import STSeparableBatchConv3d
from openretina.modules.layers.scaling import Bias3DLayer


def temporal_smoothing(sin: torch.Tensor, cos: torch.Tensor) -> torch.Tensor:
    smoother = torch.linspace(0.1, 0.9, sin.shape[2], device=sin.device)[None, None, :]
    size = float(sin.shape[0])
    reg = torch.sum((smoother * sin) ** 2) / size
    reg += torch.sum((smoother * cos) ** 2) / size
    return reg


class Core(nn.Module):
    def initialize(self) -> None:
        pass

    def regularizer(self) -> torch.Tensor | float:
        warnings.warn(
            f"Regularizer not implemented for {self.__class__.__name__}", category=RuntimeWarning, stacklevel=2
        )
        return 0.0

    def __repr__(self) -> str:
        s = f"{repr(super())} [{self.__class__.__name__} regularizers: "
        ret = []
        for attr in filter(lambda x: "gamma" in x or "skip" in x, dir(self)):
            ret.append(f"{attr} = {getattr(self, attr)}")
        return s + "|".join(ret) + "]\n"

    def save_weight_visualizations(self, folder_path: str, file_format: str = "jpg") -> None:
        print(f"Save weight visualization of {self.__class__.__name__} not implemented.")


class Core3d(Core):
    def initialize(self) -> None:
        self.apply(self.init_conv)

    @staticmethod
    def init_conv(m) -> None:
        if isinstance(m, nn.Conv3d):
            nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0)


class SimpleCoreWrapper(Core):
    def __init__(
        self,
        channels: tuple[int, ...],
        temporal_kernel_sizes: tuple[int, ...],
        spatial_kernel_sizes: tuple[int, ...],
        gamma_input: float,
        gamma_temporal: float,
        gamma_in_sparse: float,
        gamma_hidden: float,
        dropout_rate: float = 0.0,
        cut_first_n_frames: int = 30,
        maxpool_every_n_layers: int | None = None,
        downsample_input_kernel_size: tuple[int, int, int] | None = None,
        input_padding: bool = False,
        hidden_padding: bool = True,
    ):
        # Input validation
        if len(channels) < 2:
            raise ValueError(f"At least two channels required (input and output channel), {channels=}")
        if len(temporal_kernel_sizes) != len(channels) - 1:
            raise ValueError(
                f"{len(channels) - 1} layers, but only {len(temporal_kernel_sizes)} "
                f"temporal kernel sizes. {channels=} {temporal_kernel_sizes=}"
            )
        if len(temporal_kernel_sizes) != len(spatial_kernel_sizes):
            raise ValueError(
                f"Temporal and spatial kernel sizes must have the same length."
                f"{temporal_kernel_sizes=} {spatial_kernel_sizes=}"
            )

        super().__init__()
        self.gamma_input = gamma_input
        self.gamma_temporal = gamma_temporal
        self.gamma_in_sparse = gamma_in_sparse
        self.gamma_hidden = gamma_hidden
        self._cut_first_n_frames = cut_first_n_frames
        self._downsample_input_kernel_size = (
            list(downsample_input_kernel_size) if downsample_input_kernel_size is not None else None
        )
        if self._cut_first_n_frames and not input_padding:
            warnings.warn(
                (
                    "Cutting frames from the core output can lead to unexpected results if the input is not padded."
                    f"{self._cut_first_n_frames=}, {input_padding=}. Double check the core output shape."
                ),
                UserWarning,
                stacklevel=2,
            )

        self._input_weights_regularizer_spatial = FlatLaplaceL23dnorm(padding=0)

        self.features = torch.nn.Sequential()
        for layer_id, (num_in_channels, num_out_channels) in enumerate(zip(channels[:-1], channels[1:], strict=True)):
            layer: dict[str, torch.nn.Module] = OrderedDict()
            padding = "same" if input_padding and layer_id == 0 else ("same" if hidden_padding and layer_id != 0 else 0)
            layer["conv"] = STSeparableBatchConv3d(
                num_in_channels,
                num_out_channels,
                log_speed_dict={},
                temporal_kernel_size=temporal_kernel_sizes[layer_id],
                spatial_kernel_size=spatial_kernel_sizes[layer_id],
                bias=False,
                padding=padding,
            )
            layer["norm"] = torch.nn.BatchNorm3d(num_out_channels, momentum=0.1, affine=True)
            layer["bias"] = Bias3DLayer(num_out_channels)
            layer["nonlin"] = torch.nn.ELU()
            if dropout_rate > 0.0:
                layer["dropout"] = torch.nn.Dropout3d(p=dropout_rate)
            if maxpool_every_n_layers is not None and (layer_id % maxpool_every_n_layers) == 0:
                layer["pool"] = torch.nn.MaxPool3d((1, 2, 2))
            self.features.add_module(f"layer{layer_id}", torch.nn.Sequential(layer))  # type: ignore

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        if self._downsample_input_kernel_size is not None:
            input_ = torch.nn.functional.avg_pool3d(input_, kernel_size=self._downsample_input_kernel_size)  # type: ignore
        res = self.features(input_)
        # To keep compatibility with hoefling model scores
        res_cut = res[:, :, self._cut_first_n_frames :, :, :]
        return res_cut

    def spatial_laplace(self) -> torch.Tensor:
        return self._input_weights_regularizer_spatial(self.features[0].conv.weight_spatial, avg=False)

    def temporal_smoothness(self) -> torch.Tensor:
        results = [temporal_smoothing(x.conv.sin_weights, x.conv.cos_weights) for x in self.features]
        return torch.sum(torch.stack(results))

    def group_sparsity_0(self) -> torch.Tensor:
        result_array = []
        for layer in self.features:
            result = layer.conv.weight_spatial.pow(2).sum([2, 3, 4]).sqrt().sum(1) / torch.sqrt(
                1e-8 + layer.conv.weight_spatial.pow(2).sum([1, 2, 3, 4])
            )
            result_array.append(result.sum())

        return torch.sum(torch.stack(result_array))

    def group_sparsity(self) -> torch.Tensor:
        sparsities: list[torch.Tensor] = []
        for feat in self.features[1:]:
            val = feat.conv.weight_spatial.pow(2).sum([2, 3, 4]).sqrt().sum(1) / torch.sqrt(
                1e-8 + feat.conv.weight_spatial.pow(2).sum([1, 2, 3, 4])
            )
            sparsities.append(val)
        return torch.sum(torch.stack(sparsities))

    def regularizer(self) -> torch.Tensor:
        res: torch.Tensor = 0.0  # type: ignore
        for weight, reg_fn in [
            (self.gamma_input, self.spatial_laplace),
            (self.gamma_hidden, self.group_sparsity),
            (self.gamma_temporal, self.temporal_smoothness),
            (self.gamma_in_sparse, self.group_sparsity_0),
        ]:
            # lazy calculation of regularization functions
            if weight != 0.0:
                res += weight * reg_fn()
        return res

    def plot_weight_visualization(self, layer: int, in_channel: int, out_channel: int) -> plt.Figure:
        if layer >= len(self.features):
            raise ValueError(f"Requested layer {layer}, but only {len(self.features)} layers present.")
        conv_obj = self.features[layer].conv
        fig = conv_obj.plot_weights(in_channel, out_channel)
        return fig

    def save_weight_visualizations(self, folder_path: str, file_format: str = "jpg") -> None:
        for i, layer in enumerate(self.features):
            output_dir = os.path.join(folder_path, f"weights_layer_{i}")
            os.makedirs(output_dir, exist_ok=True)
            layer.conv.save_weight_visualizations(output_dir, file_format)
            print(f"Saved weight visualization at path {output_dir}")


class DummyCore(Core):
    """
    A dummy core that does nothing. Used for readout only models, like the LNP model.
    """

    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, x, data_key=None, **kwargs):
        return x

    def regularizer(self):
        return 0
